import logging
import uuid

import pytest
import geojson
import uuid
from elements.sdk.elements_sdk import ElementsSDK
from elements.sdk.tools.io import H3Tools


class TestIO:
    @pytest.mark.asyncio
    async def test_point_to_h3_boundary(self):
        h3_tools = H3Tools()
        filepath = "../resources/aois/geojson/attributed_sea_ports.geojson"
        sea_port_h3_geojson = h3_tools.points_to_h3_geometry(filename=filepath, resolution=10,
                                                             output="../resources/aois/geojson/sea_ports.geojson")
        # Need a better test here.
        assert sea_port_h3_geojson is not None

        # verify upload works
        sdk = ElementsSDK()
        aoi_collection_id = await sdk.aoi_collection.create(str(uuid.uuid4()))

        await sdk.aoi.upload(aoi_collection_id=aoi_collection_id,
                             file_path="../resources/aois/geojson/sea_ports.geojson")

        logging.info(sea_port_h3_geojson)



import codecs
import logging
from pathlib import Path

import chardet
import geojson
import h3


class H3Tools:

    def points_to_h3_geometry(self, filename, resolution: int, output: str = None, encoding: str = None):
        """
        This method takes a geojson file, a H3 resolution, and an optional file encoding type. The geojson file
        should be a Point geometry types. This will walk all points in the file, determine the H3 cell at the
        specified resolution value and return the H3 Cell geometry (polygon).

        The final result is a copy of the original file, and the Point geometry type will be replaced with "Polygon"
        geometry.

        Note: When using with TerraScope, write the file out as UTF-8 for compatibility.

        If encoding is not provided, the chardet library will be used to determine the encoding type. If the confidence
        is below 0.60, program will fail. In this case, the user must specify the encoding type. Console will provide
        the estimated encoding type, and what confidence it was determined at if the program asserts due to confidence
        threshold not being met.

        H3 Resolution Table: https://h3geo.org/docs/core-library/restable/#cell-counts
        COUNTS
        Res	Total number of cells	Number of hexagons	Number of pentagons
        0	122	                    110	                12
        1	842	                    830	                12
        2	5,882	                5,870	            12
        3	41,162	                41,150	            12
        4	288,122	                288,110	            12
        5	2,016,842	            2,016,830	        12
        6	14,117,882	            14,117,870	        12
        7	98,825,162	            98,825,150	        12
        8	691,776,122	            691,776,110	        12
        9	4,842,432,842	        4,842,432,830	    12
        10	33,897,029,882	        33,897,029,870	    12
        11	237,279,209,162	        237,279,209,150	    12
        12	1,660,954,464,122	    1,660,954,464,110	12
        13	11,626,681,248,842	    11,626,681,248,830	12
        14	81,386,768,741,882	    81,386,768,741,870	12
        15	569,707,381,193,162	    569,707,381,193,150	12

        SIZE in km^2
        Res	Average Hexagon Area (km2)	Pentagon Area* (km2)	Ratio (P/H)
        0	4,357,449.416078381	        2,562,182.162955496	    0.5880
        1	609,788.441794133	        328,434.586246469	    0.5386
        2	86,801.780398997	        44,930.898497879	    0.5176
        3	12,393.434655088	        6,315.472267516	        0.5096
        4	1,770.347654491	            896.582383141	        0.5064
        5	252.903858182	            127.785583023	        0.5053
        6	36.129062164	            18.238749548	        0.5048
        7	5.161293360	                2.604669397	            0.5047
        8	0.737327598	                0.372048038	            0.5046
        9	0.105332513	                0.053147195	            0.5046
        10	0.015047502	                0.007592318	            0.5046
        11	0.002149643	                0.001084609	            0.5046
        12	0.000307092	                0.000154944	            0.5046
        13	0.000043870	                0.000022135	            0.5046
        14	0.000006267	                0.000003162	            0.5046
        15	0.000000895	                0.000000452	            0.5046

        :param filename: str. The filepath of the .geojson file.
        :param resolution: int. A integer ranging between 0 and 15 inclusive.
        :param encoding: [Optional] str. The encoding type of the file.
        :param output: str. The filepath + filename of the file to be writen
        :return:
        """

        # Ensure resolution is within H3 resolution range.
        assert 0 <= resolution <= 15
        extension = filename.split(".")[-1]
        assert extension == "geojson"
        if encoding is None:
            encoding, confidence = self.determine_encoding(filename)
            logging.info("\nFile: {}\nEncoding: {}\n Confidence: {}".format(filename, encoding, confidence))
            assert confidence >= 0.60

        with codecs.open(filename, encoding=encoding) as f:
            file_content = geojson.load(f)

        for feature in file_content.features:
            # set memory pointer to update list in place
            point = feature.geometry.coordinates
            h3_geom = self.point_to_h3_geometry(lat=point[1], lon=point[0], resolution=resolution)
            polygon = []
            for point in h3_geom:
                polygon.append([point[0], point[1]])

            feature.geometry.type = "Polygon"
            feature.geometry.coordinates = [polygon]

        if output is not None:
            output_ext = output.split(".")[-1]
            assert output_ext == "geojson"
            with codecs.open(output, 'w', encoding="utf-8") as f:
                geojson.dump(file_content, f)

        return file_content

    @staticmethod
    def point_to_h3_geometry(lat, lon, resolution: int) -> str:
        cell_hash = h3.geo_to_h3(lat=lat, lng=lon, resolution=resolution)
        return h3.h3_to_geo_boundary(cell_hash, geo_json=True)

    @staticmethod
    def determine_encoding(filename):
        """
        Detect file encoding

        :param filename: the file to detect the encoding of.
        :return: encoding, and confidence level.

        """
        filepath = Path(filename)

        # We must read as binary (bytes) because we don't yet know encoding
        blob = filepath.read_bytes()

        encoding_detection = chardet.detect(blob)
        encoding = encoding_detection["encoding"]
        confidence = encoding_detection["confidence"]

        return encoding, confidence



from datetime import datetime

import pytest
from google.protobuf.timestamp_pb2 import Timestamp

from elements.sdk.builder.toi import TOIBuilder, TOIRuleBuilder, Frequency
from elements.sdk.elements_sdk import ElementsSDK


class TestToi:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1)])
    async def test_toi_create(self, start, finish, frequency, interval):
        sdk = ElementsSDK()
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="basic_test_toi")

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval))

        toi = await sdk.toi.create(toi_configuration.get())
        assert toi.id is not None
        assert len(toi.recurrences) == 1

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1)])
    async def test_toi_get(self, start, finish, frequency, interval):
        sdk = ElementsSDK()
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="basic_test_toi")

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval))

        toi = await sdk.toi.create(toi_configuration.get())
        assert toi.id is not None
        assert len(toi.recurrences) == 1

        tois = await sdk.toi.get([toi.id])
        assert tois is not None

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval, count", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1, 10),
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 2, 3),
        ("2019-01-05 01:00:00", "2019-02-01 01:00:00", Frequency.WEEKLY, 1, 1),
        ("2019-01-05 01:00:00", "2019-06-01 01:00:00", Frequency.MONTHLY, 1, 3)])
    async def test_toi_create_get_occurrence(self, start, finish, frequency, interval, count):
        sdk = ElementsSDK()
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="toi test occurrence {}".format(count))

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval,
                                                                     count=count))
        toi = await sdk.toi.create(toi_configuration.get())

        assert toi.id is not None
        assert len(toi.recurrences) == 1
        tois = await sdk.toi.get([toi.id])

        assert tois is not None
        for toi in tois:
            assert toi.recurrences[0].rule == "FREQ={};INTERVAL={};COUNT={}".format(Frequency(frequency).name, interval,
                                                                                    count)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval, until", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1, "20190130T010000Z"),
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 2, "20190130T010000Z"),
        ("2019-01-05 01:00:00", "2019-02-01 01:00:00", Frequency.WEEKLY, 1, "20190330T010000Z"),
        ("2019-01-05 01:00:00", "2019-06-01 01:00:00", Frequency.MONTHLY, 1, "20190630T010000Z")])
    async def test_toi_create_get_until(self, start, finish, frequency, interval, until):
        sdk = ElementsSDK()
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="toi test until {}".format(until))

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval,
                                                                     until=until))
        toi = await sdk.toi.create(toi_configuration.get())

        assert toi.id is not None
        assert len(toi.recurrences) == 1
        tois = await sdk.toi.get([toi.id])

        assert tois is not None
        for toi in tois:
            assert toi.recurrences[0].rule == "FREQ={};INTERVAL={};UNTIL={}".format(Frequency(frequency).name, interval,
                                                                                    until)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequencies, interval", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", [Frequency.DAILY, Frequency.WEEKLY], 1)])
    async def test_toi_multiple_recurrence(self, start, finish, frequencies, interval):
        sdk = ElementsSDK()
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="toi test multiple recurrence")

        for frequency in frequencies:
            toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                         interval=interval))
        toi = await sdk.toi.create(toi_configuration.get())

        assert toi.id is not None
        assert len(toi.recurrences) == 2
        tois = await sdk.toi.get([toi.id])

        assert tois is not None
        assert len(toi.recurrences) == 2


class TestToiBuilder:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1),
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 2),
        ("2019-01-05 01:00:00", "2019-02-01 01:00:00", Frequency.WEEKLY, 1),
        ("2019-01-05 01:00:00", "2019-06-01 01:00:00", Frequency.MONTHLY, 1)])
    async def test_toi_build_basic(self, start, finish, frequency, interval):
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="basic_test_toi")

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval))

        start_ts = Timestamp()
        start_ts.FromDatetime(datetime.strptime(start, datetime_format))
        finish_ts = Timestamp()
        finish_ts.FromDatetime(datetime.strptime(finish, datetime_format))

        toi = toi_configuration.get()
        assert toi.start_local == start_ts
        assert toi.finish_local == finish_ts
        assert toi.recurrences[0].rule == "FREQ={};INTERVAL={}".format(Frequency(frequency).name, interval)
        assert toi.description == "basic_test_toi"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval, count", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1, 10),
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 2, 3),
        ("2019-01-05 01:00:00", "2019-02-01 01:00:00", Frequency.WEEKLY, 1, 1),
        ("2019-01-05 01:00:00", "2019-06-01 01:00:00", Frequency.MONTHLY, 1, 3)])
    async def test_toi_build_occurrence(self, start, finish, frequency, interval, count):
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="toi test occurrence")

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval,
                                                                     count=count))

        start_ts = Timestamp()
        start_ts.FromDatetime(datetime.strptime(start, datetime_format))
        finish_ts = Timestamp()
        finish_ts.FromDatetime(datetime.strptime(finish, datetime_format))

        toi = toi_configuration.get()
        assert toi.start_local == start_ts
        assert toi.finish_local == finish_ts
        assert toi.recurrences[0].rule == "FREQ={};INTERVAL={};COUNT={}".format(Frequency(frequency).name, interval,
                                                                                count)
        assert toi.description == "toi test occurrence"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequency, interval, until", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 1, "20190130T010000Z"),
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", Frequency.DAILY, 2, "20190130T010000Z"),
        ("2019-01-05 01:00:00", "2019-02-01 01:00:00", Frequency.WEEKLY, 1, "20190230T010000Z"),
        ("2019-01-05 01:00:00", "2019-06-01 01:00:00", Frequency.MONTHLY, 1, "20190630T010000Z")])
    async def test_toi_build_until(self, start, finish, frequency, interval, until):
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="toi test until")

        toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                     interval=interval,
                                                                     until=until))

        start_ts = Timestamp()
        start_ts.FromDatetime(datetime.strptime(start, datetime_format))
        finish_ts = Timestamp()
        finish_ts.FromDatetime(datetime.strptime(finish, datetime_format))

        toi = toi_configuration.get()
        assert toi.start_local == start_ts
        assert toi.finish_local == finish_ts
        assert toi.recurrences[0].rule == "FREQ={};INTERVAL={};UNTIL={}".format(Frequency(frequency).name, interval,
                                                                                until)
        assert toi.description == "toi test until"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("start, finish, frequencies, interval", [
        ("2019-01-05 01:00:00", "2019-01-15 01:00:00", [Frequency.DAILY, Frequency.WEEKLY, Frequency.MONTHLY], 1)])
    async def test_toi_build_multiple_recurrence(self, start, finish, frequencies, interval):
        datetime_format = '%Y-%m-%d %H:%M:%S'
        toi_configuration = TOIBuilder()
        toi_configuration.build_toi(start=datetime.strptime(start, datetime_format),
                                    finish=datetime.strptime(finish, datetime_format),
                                    description="toi test until")

        for frequency in frequencies:
            toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(frequency=frequency,
                                                                         interval=interval))

        assert len(toi_configuration.get().recurrences) == len(frequencies)
        start_ts = Timestamp()
        start_ts.FromDatetime(datetime.strptime(start, datetime_format))
        finish_ts = Timestamp()
        finish_ts.FromDatetime(datetime.strptime(finish, datetime_format))

        toi = toi_configuration.get()
        assert toi.start_local == start_ts
        assert toi.finish_local == finish_ts
        for recurrence in toi.recurrences:
            found_it = False
            for frequency in frequencies:
                if recurrence.rule == "FREQ={};INTERVAL={}".format(Frequency(frequency).name, interval):
                    found_it = True
            assert found_it
        assert toi.description == "toi test until"



#!/usr/bin/env python

import click
import asyncio
import pandas as pd
import numpy as np
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup
import shapely.wkb as shp_wkb
import shapely.wkt as shp_wkt
import base64
import geopandas as gpd
import sys
import timezonefinder as tzf
import pytz


@click.command(cls=AliasedGroup, help="'imagery' command group")
@click.pass_context
def imagery(ctx):
    pass


class MatchChoice(click.Choice):
    def convert(self, value, param, ctx):
        for choice in self.choices:
            if choice.startswith(value.upper()):
                return choice
        self.fail(f"'{value}' does not match any options: {self.choices}", param, ctx)


@imagery.command('search')
@click.pass_context
@click.option('-S', '--shp_file')
@click.option('--aoi_id', default=None, help="aoi_id of geom in shp_file (default to first AOI if None)")
@click.option('-g', '--geom_wkt')
@click.option('-ts', '--datetime_start', type=pd.to_datetime)
@click.option('-te', '--datetime_end', type=pd.to_datetime)
@click.option('-d', '--data_source_id')
@click.option('-p', '--product_spec_name')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-s', '--search_service', default='SCENE', type=MatchChoice(['SCENE', 'CATALOG']))
@click.option('-o', '--out_csv', default=None)
@click.option('-q', '--quiet', is_flag=True, help="Do not print to stdout")
def imagery_search(
        ctx,
        shp_file,
        aoi_id,
        geom_wkt,
        datetime_start,
        datetime_end,
        data_source_id,
        product_spec_name,
        verbose,
        search_service='SCENE',
        out_csv=None,
        quiet=False,
):

    if geom_wkt is None:
        if shp_file is None:
            click.secho("Must provide with geom_wkt or a shp_file", fg='red')
            sys.exit(1)

        gdf = gpd.read_file(shp_file)
        if aoi_id is not None:
            if 'aoi_id' in gdf.keys():
                this_aoi = gdf['aoi_id'] == aoi_id
            elif 'Name' in gdf.keys():
                this_aoi = gdf['Name'] == aoi_id
            else:
                this_aoi = np.zeros(len(gdf)).astype(bool)
                this_aoi[0] = True
            geom = gdf[this_aoi]['geometry'].values[0]

        else:
            # if they don't give an aoi_id, just use the first 1
            geom = gdf['geometry'].values[0]

        geom_wkt = geom.wkt

    else:
        geom = shp_wkt.loads(geom_wkt)

    scenes = asyncio.run(
        wf.search_imagery(
            geom_wkt,
            datetime_start,
            datetime_end,
            data_source_id,
            product_spec_name,
            search_service=search_service,
        ))
    scenes = tsu.protobuf_to_dict(scenes)

    if not scenes:
        click.secho("No matching scenes were found.", fg='yellow')
        sys.exit()

    scenes = scenes.get('results', [])

    data = []
    tf = tzf.TimezoneFinder()
    centroid = geom.centroid
    timezone_str = tf.timezone_at(lng=centroid.x, lat=centroid.y)
    timezone = pytz.timezone(timezone_str)

    for scene in scenes:

        wkb = scene['image_geom_wkb']
        geom = shp_wkb.loads(base64.b64decode(wkb))
        centroid = geom.centroid

        utc = pd.to_datetime(scene['acquired_ts'])
        local = utc.astimezone(timezone)
        this_scene = [
            utc.strftime('%Y-%m-%dT%H:%M:%S'),
            local.strftime('%Y-%m-%dT%H:%M:%S'),
            centroid.x,
            centroid.y,
            scene['metadata'].get('aoi_coverage_percent', np.nan),
            scene['metadata'].get('cloud_cover', np.nan),
            scene['metadata'].get('sat_azimuth', np.nan),
            scene['metadata'].get('sat_elevation', np.nan),
            scene['metadata'].get('sun_azimuth', np.nan),
            scene['metadata'].get('sun_elevation', np.nan),
        ]
        if verbose:
            this_scene += [
                scene['id'],
                scene['provider_scene_id'],
                scene['image_geom_wkb'],
            ]
        data.append(this_scene)

    columns = ['acquired_utc', 'acquired_local', 'lon', 'lat', 'aoi_cov%',
               'cloud_frac', 'sat_az', 'sat_el', 'sun_az', 'sun_el']
    if verbose:
        columns += ['id', 'provider_scene_id', 'wkb']
    df = pd.DataFrame(data, columns=columns)
    if not quiet:
        click.echo(df.to_string(float_format='%.2f'))
    if out_csv:
        df.to_csv(out_csv)




import click


class AliasedBase(click.Group):
    """click.Group class to overload methods for partial command string parsing and help formatting

    """

    def get_command(self, ctx, cmd_name):
        """Overload of the get_command() method.

        It's handy to be able to type an abbreviation for commands, so
        this method overloads the get_command() method to check the
        input and see if it is an abbeviation of a command.  If it
        finds a match, it returns that.  If the input is too short and
        matches multiple commands it will fail.
        """
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)

            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((cmd, subcommand, help))

            if rows:
                cmds = []
                groups = []
                super_groups = []
                for row in rows:
                    cmd, subcmd, help = row
                    if isinstance(cmd, AliasedSuperGroup):
                        super_groups.append((subcmd, help))
                    elif isinstance(cmd, AliasedGroup):
                        groups.append((subcmd, help))
                    else:
                        cmds.append((subcmd, help))

                if super_groups:
                    with formatter.section(("Command Super-Groups")):
                        formatter.write_dl(super_groups)
                if groups:
                    with formatter.section(("Command Groups")):
                        formatter.write_dl(groups)
                if cmds:
                    with formatter.section(("Commands")):
                        formatter.write_dl(cmds)


class AliasedGroup(AliasedBase):
    """Alias for AliasedBase to be used for command groups

    All this does is provide a distinct type of command group,
    so that the formatter in AliasedBase knows what type of
    command it's dealing with.  It just allows for a clearer
    help message.
    """
    pass


class AliasedSuperGroup(AliasedBase):
    """Alias for AliasedBase to be used for command super-groups

    All this does is provide a distinct type of command group,
    so that the formatter in AliasedBase knows what type of
    command it's dealing with.  It just allows for a clearer
    help message.
    """
    pass



from typing import List

import chardet
from elements_api.models.common_models_pb2 import Pagination


class SDKSupport:
    @staticmethod
    async def get_all_paginated_objects(request, api_function, timeout) -> List:
        responses = []

        response = await api_function(request, timeout=timeout)
        responses.append(response)

        # Add subsequent result pages if exist
        pagination: Pagination = response.pagination
        while pagination.next_page_token:
            # update request to call the new page
            pagination = Pagination(
                page_token=responses[len(responses) - 1].pagination.next_page_token,
                page_size=responses[len(responses) - 1].pagination.page_size
            )
            request.pagination.MergeFrom(pagination)
            responses.append(await api_function(request, timeout=timeout))
            # set the next page token now that we have the new page
            pagination = Pagination(next_page_token=responses[len(responses) - 1].pagination.next_page_token)

        return responses

    @staticmethod
    def determine_encoding(filename):
        with open(filename, 'rb') as f:
            file_bytes = f.read()
            detection = chardet.detect(file_bytes)
            encoding = detection["encoding"]
            confidence = detection["confidence"]

        return encoding, confidence



import pytest

from elements.sdk.elements_sdk import ElementsSDK


class TestDataSource:
    @pytest.mark.asyncio
    async def test_get(self):
        sdk = ElementsSDK()
        ids = ["blogwatcher_pings", "adsbx_pings"]
        sources = await sdk.data_source.get(ids=ids)
        assert len(sources) == len(ids)

    @pytest.mark.asyncio
    async def test_list(self):
        sdk = ElementsSDK()
        data_sources = await sdk.data_source.list()
        assert len(data_sources) > 0

        data_sources = await sdk.data_source.list(search_text="exact-earth")
        assert len(data_sources) == 2


class TestDataType:
    @pytest.mark.asyncio
    async def test_get(self):
        sdk = ElementsSDK()
        data_types = await sdk.data_type.get(ids=["pings"])
        assert len(data_types) == 1

    @pytest.mark.asyncio
    async def test_list(self):
        sdk = ElementsSDK()
        data_types = await sdk.data_type.list()
        assert len(data_types) > 0

        data_types = await sdk.data_type.list(search_text="ping")
        assert len(data_types) == 2

    @pytest.mark.asyncio
    async def test_create(self):
        sdk = ElementsSDK()
        data_type = await sdk.data_type.create(name="test_data_type", description="test description", schema="", data_source_ids=[])



#!/usr/bin/env python

import click
import asyncio
import os
import json
import yaml
import pandas as pd
import tabulate
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup
import elements.cli.lib.workflow as wf
from google.protobuf.json_format import MessageToDict


@click.command(cls=AliasedGroup, help="Analysis 'version' command group")
@click.pass_context
def version(ctx):
    pass


@version.command('get')
@click.pass_context
@click.option('-iV', '--analysis_version_id', type=str)
@click.option('-iA', '--analysis_id', type=str)
@click.option('-nA', '--analysis_name', type=str)
@click.option('-C', '--only_columns', multiple=True)
@click.option('-c', '--append_columns', multiple=True)
@click.option('-T', '--transpose', is_flag=True)
@click.option('-om', '--output_manifest', type=str,
              help="Write manifest to this filename.", default=None)
def analysis_version_get(
        ctx,
        analysis_version_id,
        analysis_id,
        analysis_name,
        only_columns,
        append_columns,
        transpose,
        output_manifest=None,
):

    columns = ['id', 'created_on', 'analysis.name', 'analysis.author']
    if only_columns:
        columns = list(only_columns)
    if append_columns:
        columns = columns + list(append_columns)

    # if they provide an explicit algo version ID, get it.
    analysis_versions = []
    if analysis_version_id:
        analysis_versions += asyncio.run(wf.get_analysis_versions(analysis_version_id=analysis_version_id))

    analysis_ids = []

    # If they give us a uuid, use it.  Otherwise, assume it's a name and look it up.
    if analysis_id:
        analysis_ids = [analysis_id]
    if analysis_name:
        analyses = asyncio.run(wf.list_analyses(analysis_name))
        analysis_ids = [analysis.id for analysis in analyses]

    for analysis_id in analysis_ids:
        analysis_versions += asyncio.run(wf.get_analysis_versions(analysis_id=analysis_id))

    if analysis_versions:
        data = []
        for analysis_version in analysis_versions:
            analysis_version_dict = tsu.protobuf_to_dict(analysis_version)
            data.append(analysis_version_dict)

        df = pd.DataFrame(data=data)
        if 'all' not in columns:
            df = df[tsu.match_columns(df.columns, columns)]
        df = df.rename(columns={'id': 'analysis_version_id'})
        if transpose:
            dfT = df.T
            if 'algorithm_versions' in df.keys():
                algo_ver = df['algorithm_versions'].values[0]
                dfT.drop('algorithm_versions', inplace=True)
                dfT = dfT.append(pd.json_normalize(algo_ver).T)
            df = tabulate.tabulate(dfT, maxcolwidths=[None, 50])

        tsu.set_pandas_display()
        click.secho(f"analysis_id: {analysis_id}", fg='cyan')
        click.echo(df)

    if output_manifest:

        n_versions = len(analysis_versions)
        for analysis_version in analysis_versions:

            manifest = MessageToDict(analysis_version.analysis_manifest)
            analysis_id = analysis_version.analysis.id
            manifest['analysis_id'] = analysis_id
            manifest['analysis_name'] = analysis_version.analysis.name

            output_manifest_fname = output_manifest
            if n_versions > 1:
                fname, ext = os.path.splitext(output_manifest)
                output_manifest_fname = f"{fname}_{analysis_id[:8]}.{ext}"
            with open(output_manifest_fname, 'w') as fp:
                if output_manifest.endswith('json'):
                    json.dump(manifest, fp, indent=4)
                else:
                    yaml.dump(manifest, fp)


@version.command('create')
@click.pass_context
@click.option('-iA', '--analysis_id', type=str)
@click.option('-iv', '--algorithm_version_id', type=str)
@click.option('-nA', '--analysis_name', type=str)
@click.option('-D', '--description', type=str)
@click.option('-V', '--version', type=str)
@click.option('-T', '--tags', type=str, multiple=True)
@click.option('--dry_run', is_flag=True, default=False)
def analysis_version_create(
        ctx,
        analysis_id,
        algorithm_version_id,
        analysis_name,
        description,
        version,
        tags,
        dry_run,
):

    wf.check_environment_complete(raise_on_failure=True)
    manifest = wf.create_analysis_manifest(
        analysis_name,
        algorithm_version_id,
        description,
        version,
        tags
    )
    if not dry_run:
        analysis_version_id = asyncio.run(
            wf.create_analysis_version(
                analysis_id,
                manifest,
            )
        )

        click.echo(f"Your new analysis_version_id is: {analysis_version_id}")



from typing import List

from google.protobuf.timestamp_pb2 import Timestamp
from elements_api import ElementsAsyncClient
from elements_api.models.analysis_computation_pb2 import (
    AnalysisComputationCreateRequest, AnalysisComputationRunRequest, AnalysisComputation, AnalysisComputationGetRequest,
    AnalysisComputationListRequest
)
from elements_api.models.analysis_config_pb2 import (
    AnalysisConfigCreateRequest, AnalysisConfigGetRequest, AnalysisConfigListRequest, AnalysisConfigDeactivateRequest,
    AnalysisConfig
)
from elements_api.models.analysis_pb2 import (
    Analysis, AnalysisCreateRequest, AnalysisGetRequest, AnalysisListRequest
)
from elements_api.models.analysis_version_pb2 import (
    AnalysisVersionListRequest, AnalysisVersionGetRequest, AnalysisVersionCreateRequest, AnalysisVersion
)
from elements_api.models.common_models_pb2 import Pagination

from elements.sdk.builder.analysis import AnalysisManifest, AnalysisConfiguration
from elements.sdk.tools.sdk_support import SDKSupport



ANALYSIS_DEFAULT_PAGE_SIZE = 200


class APIAnalysis:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, name: str, author: str) -> Analysis:
        """
        Create a top level analysis object. This object is lightweight and is used to connect all versions of an
        analysis together. Once an analysis has been created, an analysis_version must be created using a manifest.
        Next an analysis_config may be created to specify a particular configuration of an analysis version,
        which can be used in an analysis_computation.
        :param name:
        :param author:
        :return: Analysis
        """

        request = AnalysisCreateRequest(
            name=name,
            author=author
        )
        response = await self.__client.api.analysis.create(request, timeout=self.__timeout)

        return response.analysis

    async def update(self, analysis_id: str, name: str = None, author: str = None):
        """
        Update the mutable fields of a top level analysis. This should be done rarely and with caution as it will
        affect all analysis_versions, and analysis_configs associated with this analysis.
        :return:
        """
        pass

    async def get(self, ids: List, pagination: Pagination = None) -> List[Analysis]:
        """
        Retrieve the metadata of a set of analyses.

        :param pagination:
        :param ids: List of analysis ids to get the metadata for.
        :return: List[Analysis]
        """
        if pagination is None:
            Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE,
            )
        request = AnalysisGetRequest(
            ids=ids,
            pagination=pagination
        )
        sdk_support = SDKSupport()
        responses = await sdk_support.get_all_paginated_objects(request,
                                                                api_function=self.__client.api.analysis.get,
                                                                timeout=self.__timeout)
        analyses = []
        for response in responses:
            analyses.extend(response.analyses)
        return analyses

    async def list(self, **kwargs) -> List[Analysis]:
        """
        Retrieve the metadata for all analyses that this user has access to and that match the provided search filters,
        if any.

        Empty Arg list will list all available Analyses.

        :param kwargs
            - search_text: str
            - min_created_on: Datetime
            - max_created_on: Datetime
            - pagination: Pagination
        :return: List[Analysis]
        """
        request_fragments = []
        fetch_all = False
        if 'pagination' not in kwargs.keys():
            request_fragments.append((AnalysisListRequest(pagination=Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            ))))
            fetch_all = True

        if 'raw_response' not in kwargs.keys():
            kwargs['raw_response'] = False

        for key in kwargs.keys():
            if key == 'search_text':
                assert isinstance(kwargs[key], str)
                request_fragments.append(AnalysisListRequest(search_text=kwargs[key]))
            if key == 'min_created_on':
                min_created_on = Timestamp()
                min_created_on.FromDatetime(kwargs[key])
                request_fragments.append(AnalysisListRequest(min_created_on=min_created_on))
            if key == 'max_created_on':
                max_created_on = Timestamp()
                max_created_on.FromDatetime(kwargs[key])
                request_fragments.append(AnalysisListRequest(min_created_on=max_created_on))
            if key == 'pagination':
                request_fragments.append(AnalysisListRequest(pagination=kwargs[key]))

        request = AnalysisListRequest()
        for param in request_fragments:
            request.MergeFrom(param)

        if fetch_all:
            sdk_support = SDKSupport()
            responses = await sdk_support.get_all_paginated_objects(request,
                                                                    api_function=self.__client.api.analysis.list,
                                                                    timeout=self.__timeout)
            if kwargs['raw_response']:
                return responses
            else:
                analyses = []
                for response in responses:
                    analyses.extend(response.analyses)

                return analyses
        else:
            response = await self.__client.api.analysis.list(request, timeout=self.__timeout)

            if kwargs['raw_response']:
                return response
            else:
                return response.analyses


class APIAnalysisVersion:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, analysis_id: str, analysis_manifest: AnalysisManifest) -> AnalysisVersion:
        """
        Given a top level analysis, create a new version. Once an analysis version has been created, it is immutable. Any changes require a new version to be
        created. The purpose of this is to either upgrade the structure of the Analysis DAG OR update the versions of
        the algorithms assigned to the Analysis.

        :return: AnalysisVersion
        """
        request = AnalysisVersionCreateRequest(
            analysis_id=analysis_id,
            analysis_manifest=analysis_manifest.get()
        )
        response = await self.__client.api.analysis_version.create(request, timeout=self.__timeout)
        return response.analysis_version

    async def get(self, ids: List,
                  include_manifest: bool = False,
                  include_algorithm_details: bool = False,
                  pagination: Pagination = None) -> List[AnalysisVersion]:
        """
        Get the details for a specific analysis_version. By default it does not include the analysis_manifest.
        Setting include_manifest to true, will also include the manifest in the response for each
        requested analysis_version.

        :param: pagination

        :return: List[AnalysisVersion]
        """
        if pagination is None:
            pagination = Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            )
        request = AnalysisVersionGetRequest(
            ids=ids,
            include_manifest=include_manifest,
            include_algorithm_details=include_algorithm_details,
            pagination=pagination
        )
        sdk_support = SDKSupport()
        responses = await sdk_support.get_all_paginated_objects(request,
                                                                api_function=self.__client.api.analysis_version.get,
                                                                timeout=self.__timeout)
        analysis_versions = []
        for response in responses:
            analysis_versions.extend(response.analysis_versions)
        return analysis_versions

    async def list(self, **kwargs) -> List[AnalysisVersion]:
        """
        List all the analysis versions that the user has access to. Search text will be applied to the entire manifest
        returning all analysis_versions whose manifest contains the search text. Tags will be matched exactly with the
        tags on the manifest. By default, this method only returns the semantically latest version of each analysis that
        has not been deprecated, It does not include manifests in the response. If include_all_versions is set to true,
        all prior versions of an analysis including deprecated analysis versions will be returned, as well. If
        include_manifest is set to true, the corresponding manifest of each analysis_version will also be returned.

        :param kwargs:
            - analysis_id: str
            - search_text: str
            - tag: str
            - min_created_on: Datetime
            - max_created_on: Datetime
            - include_all_versions: bool
            - pagination: Pagination
        :return: List[AnalysisVersion]
        """

        assert kwargs.keys() is not None

        request_fragments = []
        fetch_all = False
        if 'pagination' not in kwargs.keys():
            request_fragments.append((AnalysisVersionListRequest(pagination=Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            ))))
            fetch_all = True

        if 'raw_response' not in kwargs.keys():
            kwargs['raw_response'] = False

        if 'include_manifest' in kwargs.keys():
            request_fragments.append(AnalysisVersionListRequest(include_manifest=True))

        if 'include_all_versions' in kwargs.keys():
            request_fragments.append(AnalysisVersionListRequest(include_all_versions=True))

        for key in kwargs.keys():
            if key == 'analysis_id':
                request_fragments.append(AnalysisVersionListRequest(analysis_id=kwargs[key]))
            if key == 'search_text':
                request_fragments.append(AnalysisVersionListRequest(search_text=kwargs[key]))
            if key == 'tag':
                request_fragments.append(AnalysisVersionListRequest(tag=kwargs[key]))
            if key == 'min_created_on':
                min_created_on = Timestamp()
                min_created_on.FromDatetime(kwargs[key])
                request_fragments.append(AnalysisVersionListRequest(min_created_on=min_created_on))
            if key == 'max_created_on':
                max_created_on = Timestamp()
                max_created_on.FromDatetime(kwargs[key])
                request_fragments.append(AnalysisVersionListRequest(max_created_on=max_created_on))
            if key == 'include_all_versions':
                request_fragments.append(AnalysisVersionListRequest(include_all_versions=kwargs[key]))
            if key == 'pagination':
                request_fragments.append(AnalysisVersionListRequest(pagination=kwargs[key]))

        request = AnalysisVersionListRequest()
        for request_fragment in request_fragments:
            request.MergeFrom(request_fragment)

        if fetch_all:
            sdk_support = SDKSupport()
            responses = await sdk_support.get_all_paginated_objects(request,
                                                                    api_function=self.__client.api.analysis_version.list,
                                                                    timeout=self.__timeout)
            if kwargs['raw_response']:
                return responses
            else:
                analysis_versions = []
                for response in responses:
                    analysis_versions.extend(response.analysis_versions)
                return analysis_versions
        else:
            response = await self.__client.api.analysis_version.list(request, timeout=self.__timeout)

            if kwargs['raw_response']:
                return response
            else:
                return response.analysis_versions

    async def deprecate(self, analysis_version_ids: List):
        """
        Deprecate the specific version of this analysis. Deprecated analysiss can still be searched for, metadata can
        be retrieved, and can still be used to create new analysis_configs which can be used as part of
        analysis_computations, but they are no longer actively supported. All analysis_configs created from this
        analysis_version are also deprecated, but can still be used in new analysis_computations.
        :param analysis_version_ids:
        :return:
        """
        pass

    async def deactivate(self, analysis_version_ids: List):
        """
        Deactivate the specific version of this analysis. Metadata for deactivated analyses can still be retrieved
        directly, but they will no longer show up in search results from the list endpoint.Deactivated
        analysis_versions cannot be used to create new analysis_configs. All analysis_configs created from this
        analysis_version are also deactivated and can no longer be used to create new analysis_computations.
        Existing computations that are already using analysis_configs created from this analysis_version
        will still continue to run.
        :param analysis_version_ids:
        :return:
        """
        pass


class APIAnalysisConfig:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, analysis_version_id: str,
                     name: str,
                     description: str,
                     algorithm_config_nodes: List) -> AnalysisConfig:
        """
        Creates a configuration of the specified analysis version, including concrete values for all of the analysis
        parameters. This configuration is stored in the system and the analysis_config is returned to the user.

        :param analysis_version_id:
        :param name
        :param description:
        :param algorithm_config_nodes:

        :return: AnalysisConfig
        """
        request = AnalysisConfigCreateRequest(
            analysis_version_id=analysis_version_id,
            name=name,
            description=description,
            algorithm_config_nodes=algorithm_config_nodes
        )
        response = await self.__client.api.analysis_config.create(request, timeout=self.__timeout)

        return response.analysis_config

    async def update(self, analysis_config_id: str, analysis_config: AnalysisConfiguration):
        """
        Update details of an analysis_config, so long as it has not been locked. Once the analysis_config is used in a
        computation, it is permanently locked and the specific parameter values can no longer be updated.
        :param analysis_config_id:
        :param analysis_config:
        :return:
        """
        pass

    async def get(self, ids: List,
                  include_algorithm_details: bool = False,
                  pagination: Pagination = None) -> List[AnalysisConfig]:
        """
        Retrieve the metadata of a particular analysis configuration.

        :return: List[AnalysisConfig]
        """
        if pagination is None:
            pagination = Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            )
        request = AnalysisConfigGetRequest(
            ids=ids,
            include_algorithm_details=include_algorithm_details,
            pagination=pagination
        )

        sdk_support = SDKSupport()
        responses = await sdk_support.get_all_paginated_objects(request,
                                                                api_function=self.__client.api.analysis_config.get,
                                                                timeout=self.__timeout)
        analysis_configs = []
        for response in responses:
            analysis_configs.extend(response.analysis_configs)

        return analysis_configs

    async def list(self, **kwargs) -> List[AnalysisConfig]:
        """
        Retrieve the metadata for all analysis configurations that this user has access to and that match the provided
        search filters. By default, deactivated analysis_configs are not returned. Setting include_deactivated to true
        will also return any analysis_configs that had been deactivated.

        :param kwargs:
            - analysis_id: str
            - analysis_version_id: str
            - subject_id: str
            - search_text: str
            - tag: str
            - min_created_on: Datetime
            - max_created_on: Datetime
            - include_algorithm_details: bool
            - include_all_versions: bool
            - pagination: Pagination

        :return: List[AnalysisConfig]
        """
        assert kwargs.keys() is not None

        request_fragments = []
        fetch_all = False
        if 'pagination' not in kwargs.keys():
            request_fragments.append(AnalysisConfigListRequest(pagination=Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            )))
            fetch_all = True

        if 'raw_response' not in kwargs.keys():
            kwargs['raw_response'] = False

        for key in kwargs.keys():
            if key == 'analysis_id':
                request_fragments.append(AnalysisConfigListRequest(analysis_id=kwargs[key]))
            if key == 'analysis_version_id':
                request_fragments.append(AnalysisConfigListRequest(analysis_version_id=kwargs[key]))
            if key == 'subject_id':
                request_fragments.append(AnalysisConfigListRequest(subject_id=kwargs[key]))
            if key == 'search_text':
                request_fragments.append(AnalysisConfigListRequest(search_text=kwargs[key]))
            if key == 'tag':
                request_fragments.append(AnalysisConfigListRequest(tag=kwargs[key]))
            if key == 'min_created_on':
                min_created_on = Timestamp()
                min_created_on.FromDatetime(kwargs[key])
                request_fragments.append(AnalysisConfigListRequest(min_created_on=min_created_on))
            if key == 'max_created_on':
                max_created_on = Timestamp()
                max_created_on.FromDatetime(kwargs[key])
                request_fragments.append(AnalysisConfigListRequest(max_created_on=max_created_on))
            if key == 'pagination':
                request_fragments.append(AnalysisConfigListRequest(pagination=kwargs[key]))

        if 'include_deactivated' in kwargs.keys():
            request_fragments.append(AnalysisConfigListRequest(include_deactivated=kwargs['include_deactivated']))
        else:
            request_fragments.append(AnalysisConfigListRequest(include_deactivated=False))

        # uncomment this when include_all_versions is available
        # if 'include_all_versions' in kwargs.keys():
        #    request_fragments.append(AnalysisConfigListRequest(include_all_versions=kwargs['include_all_versions']))
        # else:
        #    request_fragments.append(AnalysisConfigListRequest(include_all_versions=False))

        request = AnalysisConfigListRequest()
        for request_fragment in request_fragments:
            request.MergeFrom(request_fragment)

        if fetch_all:
            sdk_support = SDKSupport()
            responses = await sdk_support.get_all_paginated_objects(request,
                                                                    api_function=self.__client.api.analysis_config.list,
                                                                    timeout=self.__timeout)
            if kwargs['raw_response']:
                return responses
            else:
                analysis_configs = []
                for response in responses:
                    analysis_configs.extend(response.analysis_configs)
                return analysis_configs
        else:
            response = await self.__client.api.analysis_config.list(request)

            if kwargs['raw_response']:
                return response
            else:
                return response.analysis_configs

    async def deprecate(self, analysis_config_ids: List):
        """
        Deprecate the specific configuration of an analysis_version. Deprecated analysis_configs can still be searched
        for, metadata can be retrieved, and can still be used in new analysis_computations, but they are no longer
        actively supported.
        :param analysis_config_ids:
        :return:
        """
        pass

    async def deactivate(self, analysis_config_ids: List[str]):
        """
        Deactivate the specific configuration of an analysis_version. Deactivated analysis_configs will no longer
        show up in searches, but metadata can still be retrieved for deactivated analysis_configs via the get endpoint.
        Deactivated analysis_configs cannot be used in any new analysis_computations, but existing analysis_computations
         will continue to function properly.
        :param analysis_config_ids:
        :return:
        """
        analysis_config_deactivate_request = AnalysisConfigDeactivateRequest(
            ids=analysis_config_ids
        )
        analysis_config_deactivate_response = await self.__client.api.analysis_config.deactivate(
            analysis_config_deactivate_request, timeout=self.__timeout)

        return analysis_config_deactivate_response

    async def delete(self, analysis_config_ids: List):
        """
        Delete the specified analysis_config, if and only if it has not been used to create an analysis_computation.
        Once an analysis_config has been used to create an analysis_computation it can never be deleted.
        The user must have permission to access the specified analysis.
        :param analysis_config_ids:
        :return:
        """
        pass


class APIAnalysisComputation:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, analysis_config_id: str,
                     aoi_collection_id: str,
                     toi_id: str) -> AnalysisComputation:
        """
        Create the runnable component in the system, by specifiying the ANALYSIS to be run, the set of AOIs to be
        run on, and the TOI to specify the recurrence. The analysis_computation_id is returned so that it can be run.
        Note that until an analysis_computation has been run, its underlying .resources are NOT locked and therefore can
        be changed.
        :param analysis_config_id:
        :param aoi_collection_id:
        :param toi_id:
        :return: AnalysisComputation object
        """

        request = AnalysisComputationCreateRequest(
            analysis_config_id=analysis_config_id,
            toi_id=toi_id,
            aoi_collection_id=aoi_collection_id
        )
        response = await self.__client.api.analysis_computation.create(request, timeout=self.__timeout)
        return response.analysis_computation

    async def run(self, analysis_computation_ids: List):
        """
        The first time this endpoint is called, the system initiates the running of the provided
        analysis_computation_id. This causes credits to be place in pending status and eventually consumed when the
        computation completes. Credits are consumed as each result is produced. Once an analysis has been run, all of
        its underlying constituent parts are considered locked and cannot be modified.

        :param analysis_computation_ids:
        :return: AnalysisSubmitResponse
        """

        request = AnalysisComputationRunRequest(
            ids=analysis_computation_ids
        )
        await self.__client.api.analysis_computation.run(request, timeout=self.__timeout)

    async def get(self, ids: List, pagination: Pagination = None) -> List[AnalysisComputation]:
        """
        Retrieves the details of the provided analysis_computation including its current state and its progress.

        **NOT_STARTED**: Analysis computations that have been created but not yet passed into the "run" endpoint.
        **IN_PROGRESS**: Analysis computations that have been "run" but are not yet complete.
        **COMPLETE**: Once an analysis computation has run on all AOIs and for all TOIs and ALGOs and all
        results are either complete or failed.

        :return: List[AnalysisComputation]
        """
        if pagination is None:
            pagination = Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            )

        request = AnalysisComputationGetRequest(
            ids=ids,
            pagination=pagination
        )
        sdk_support = SDKSupport()
        responses = await sdk_support.get_all_paginated_objects(request,
                                                                api_function=self.__client.api.analysis_computation.get,
                                                                timeout=self.__timeout)
        analysis_computations = []
        for response in responses:
            analysis_computations.extend(response.analysis_computations)

        return analysis_computations

    async def update(self, analysis_computation_id: str, **kwargs):
        """
        Update an analysis computation's aoi, toi, algorithm config id. The analysis computation must be unlocked.
        :param analysis_computation_id:
        :param kwargs:
            - analysis_config_id: str
            - id: str,
            - toi_id: str
        :return:
        """
        pass

    async def list(self, **kwargs) -> List[AnalysisComputation]:
        """
        "Lists the details for all analysis_computations that the requester has access to including
        their current state and their progress.

        **NOT_STARTED**: Analysis computations that have been created but not yet passed into the ""run"" endpoint.
        **IN_PROGRESS**: Analysis computations that have been ""run"" but are not yet complete.
        **COMPLETE**: Once an analysis computation has run on all AOIs and for all TOIs and ALGOs
            and all results are either complete or failed."
        :param kwargs:
            - state: state_enum,
            - status: status_enum,
            - max_created_on: Datetime,
            - min_created_on: Datetime,
            - analysis_config_id: str,
            - toi_id: str,
            - id: str,
            - analysis_id: str,
            - pagination: str
        :return: List[AnalysisComputation]
        """
        fragments = []
        if 'pagination' not in kwargs.keys():
            fragments.append(AnalysisComputationListRequest(pagination=Pagination(
                page_size=ANALYSIS_DEFAULT_PAGE_SIZE
            )))

        for key in kwargs.keys():
            if key == 'state':
                fragments.append(AnalysisComputationListRequest(state=kwargs[key]))
            if key == 'status':
                fragments.append(AnalysisComputationListRequest(status=kwargs[key]))
            if key == 'min_created_on':
                min_created_on = Timestamp()
                min_created_on.FromDatetime(kwargs[key])
                fragments.append(AnalysisComputationListRequest(submitted_min=min_created_on))
            if key == 'max_created_on':
                max_created_on = Timestamp()
                max_created_on.FromDatetime(kwargs[key])
                fragments.append(AnalysisComputationListRequest(submitted_max=max_created_on))
            if key == 'analysis_config_id':
                fragments.append(AnalysisComputationListRequest(analysis_config_id=kwargs[key]))
            if key == 'toi_id':
                fragments.append(AnalysisComputationListRequest(toi_id=kwargs[key]))
            if key == 'id':
                fragments.append(AnalysisComputationListRequest(aoi_collection_id=kwargs[key]))
            if key == 'analysis_id':
                fragments.append(AnalysisComputationListRequest(analysis_id=kwargs[key]))
            if key == 'pagination':
                fragments.append(AnalysisComputationListRequest(search_text=kwargs[key]))

        request = AnalysisComputationListRequest()
        for param in fragments:
            request.MergeFrom(param)

        sdk_support = SDKSupport()
        responses = await sdk_support.get_all_paginated_objects(
            request,
            api_function=self.__client.api.analysis_computation.list,
            timeout=self.__timeout)

        analysis_computations = []
        for response in responses:
            analysis_computations.extend(response.analysis_computations)

        return analysis_computations

    async def delete(self, analysis_computation_ids: List):
        """
        Deletes the specified analysis_computation, provided that it has not already been run. This is a permanent
        action. Must be the user that created the analysis_computation.
        :param analysis_computation_ids:
        :return:
        """
        pass

    async def pause(self, analysis_computation_ids: List):
        """
        Update the analysis_computation to move out of the *IN_PROGRESS* state. This will ensure that it no longer runs,
        until it is resumed. Whatever is currently running, will still complete.
        :param analysis_computation_ids:
        :return:
        """
        pass

    async def resume(self, analysis_computation_ids: List):
        """
        Update the analysis_computation to move back into the IN_PROGRESS state, and will again be set to run.
        :param analysis_computation_ids:
        :return:
        """
        pass

    async def history(self, analysis_computation_id: str):
        """
        Returns the history of the analysis computation including the status of all previous executions.
        :param analysis_computation_id:
        :return:
        """
        pass



#!/usr/bin/env python

import pandas as pd
import click
import re
import os
from google.protobuf.json_format import MessageToDict


class FakeDataSource:

    def __init__(self, value):
        self.value = value


def set_pandas_display():
    """Reset the default display parameters for printing pandas dataframes.

    By default pandas chooses very small column widths and shows only
    a few rows, etc.  The TS cli prints information in dataframes in
    some cases, so we'll reset these parameters to more appropriate
    values.
    """
    pd.set_option('display.width', 220)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_colwidth', 120)


def protobuf_to_dict(obj):
    """Helper Function to take a protobuf object and return a Python dict"""

    info = MessageToDict(obj, preserving_proto_field_name=True)
    info = pd.json_normalize(info, sep='.').to_dict(orient='records')
    if isinstance(info, list) and len(info) > 0:
        info = info[0]
    return info


def match_columns(columns, exprs):
    """
    """
    keep = []
    for col in columns:
        if col in keep:
            continue
        for expr in exprs:
            if col in keep:
                continue
            if re.match(expr, col):
                keep.append(col)
    return keep


def chunk_array(array, n):
    """Yield successive n-sized chunks from list 'array'."""
    for i in range(0, len(array), n):
        yield array[i:i + n].copy()


def get_author(author, raise_on_failure=False):
    """Read author from an environment variable, or return as-is.

    If the user sets author to 'ELEMENTS_AUTHOR', we'll read that
    variable and return its value.  Otherwise, we'll assume the user
    has already set it and just return what we were given.
    """

    env_var = 'ELEMENTS_AUTHOR'
    if author == env_var:
        author = os.environ.get(author, None)
        if raise_on_failure and author is None:
            raise RuntimeError(f"Environment variable '{env_var}' is not set")

    return author


def override_manifest_param(name, value, manifest_dict, do_overwrite=False):
    """
    """

    # If the user provided a value, use it, otherwise get it from the manifest.
    if value is None:

        value = manifest_dict.get(name, None)
        if value is None:
            raise ValueError(f"Must provide parameter '{name}' in the manifest, or on the command line")

    if do_overwrite:
        manifest_dict[name] = value

    return value


def increment_version(current_version, level='patch'):

    major, minor, patch = [int(v) for v in current_version.split('.')]
    if level == 'patch':
        patch += 1
    elif level == 'minor':
        minor += 1
        patch = 0
    elif level == 'major':
        major += 1
        minor = 0
        patch = 0

    new_version = '.'.join([str(major), str(minor), str(patch)])
    return new_version


def set_version(version_input, manifest, do_overwrite=False):
    """
    """
    manifest_version = manifest['metadata']['version']

    if version_input:
        if version_input in ['patch', 'minor', 'major']:
            new_version = increment_version(manifest_version, level=version_input)
        else:
            new_version = version_input
    else:
        new_version = manifest_version

    if do_overwrite:
        manifest['metadata']['version'] = new_version

    return new_version


def echo_highlight_suffix(message, suffix, color):

    click.echo(message, nl=False)
    click.secho(f"{suffix}", fg=color)



#!/usr/bin/env python

import click
import asyncio
import pandas as pd
import yaml
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup
import elements.cli.lib.workflow as wf


@click.group(cls=AliasedGroup, help="Algorithm 'config' command group")
@click.pass_context
def config(ctx):
    pass


@config.command('get')
@click.pass_context
@click.option('-ic', '--algorithm_config_id', type=str, default=None,
              help="UUID of the algorithm config to look up.")
@click.option('-iv', '--algorithm_version_id', type=str, default=None,
              help="UUID of an algorithm version to look up associated algorithm configs.")
@click.option('-n', '--algorithm_name', type=str, default=None,
              help="Name of the algorithm to look up.  All matching algo versions will be queried")
@click.option('-C', '--only_columns', multiple=True,
              help="Include only the listed columns in output.")
@click.option('-c', '--append_columns', multiple=True,
              help="Append these columns to those shown by default.")
@click.option('-T', '--transpose', is_flag=True,
              help="Display the result in transpose (helpful to show algorithms with long string values).")
def algorithm_config_get(
        ctx,
        algorithm_config_id,
        algorithm_version_id,
        algorithm_name,
        only_columns,
        append_columns,
        transpose,
):
    algorithm_version_ids = []
    if algorithm_version_id is not None:
        algorithm_version_ids.append([algorithm_version_id, None, None])

    # if requested by name, we'll look up the available version ids and run all of them.
    if algorithm_name is not None:
        algos = asyncio.run(wf.list_algorithms(algorithm_name))
        for algo in algos:
            algorithm_versions = asyncio.run(wf.get_algorithm_versions(algo.id))
            for algorithm_version in algorithm_versions:
                algorithm_version_ids.append([algorithm_version.id, algo.name, algo.id])

    for algorithm_version_info in algorithm_version_ids:
        algorithm_version_id = algorithm_version_info[0]
        algorithm_configs = asyncio.run(wf.get_algorithm_configs(algorithm_version_id=algorithm_version_id))
        _print_algorithm_configs(algorithm_version_info, algorithm_configs, only_columns, append_columns, transpose)

    if algorithm_config_id:
        algorithm_configs = asyncio.run(wf.get_algorithm_configs(algorithm_config_id=algorithm_config_id))
        algorithm_version_info = [algorithm_version_id, None, None]
        _print_algorithm_configs(algorithm_version_info, algorithm_configs, only_columns, append_columns, transpose)


def _print_algorithm_configs(
        algorithm_version_info,
        algorithm_configs,
        only_columns,
        append_columns,
        transpose
):

    algo_version_id, algo_name, algo_id = algorithm_version_info
    click.secho(f"name: {algo_name}", fg='magenta')
    click.secho(f"      algo_id:         {algo_id}", fg='magenta')
    click.secho(f"      algo_version_id: {algo_version_id}", fg='magenta')

    columns = ['name', 'id', 'description', 'created_on']
    if only_columns:
        columns = list(only_columns)
    if append_columns:
        columns = columns + list(append_columns)

    algorithm_configs = tsu.protobuf_to_dict(algorithm_configs[0])
    if 'algorithm_configs' in algorithm_configs:
        algorithm_configs = algorithm_configs['algorithm_configs']
    else:
        algorithm_configs = [algorithm_configs]

    data = []
    for algorithm_config in algorithm_configs:
        data.append(algorithm_config)
    df = pd.DataFrame(data)
    if 'all' not in columns:
        df = df[tsu.match_columns(df.columns, columns)]
    if transpose:
        df = df.T.rename(columns={0: ""})
    df = df.rename(columns={'id': 'algorithm_config_id'})
    tsu.set_pandas_display()
    click.secho(f"algorithm_version_id: {algo_version_id}", fg='cyan')
    click.echo(df)


@config.command('create')
@click.pass_context
@click.option('-iv', '--algorithm_version_id', type=str)
@click.option('-m', '--manifest_yaml', type=str)
@click.option('-nc', '--config_name', type=str)
@click.option('-d', '--config_desc', type=str)
@click.option('-s', '--data_source', type=str)
@click.option('-t', '--data_type', type=str)
@click.option('-p', '--image_processing_spec', type=str, default=None)
@click.option('--dry_run', is_flag=True)
def algorithm_config_create(
        ctx,
        algorithm_version_id,
        manifest_yaml,
        config_name,
        config_desc,
        data_source,
        data_type,
        image_processing_spec=None,
        dry_run=False,
):

    wf.check_environment_complete(raise_on_failure=True)
    input_manifest = None
    with open(manifest_yaml, 'r') as fp:
        input_manifest = yaml.safe_load(fp)

    if not data_source:
        try:
            data_source = input_manifest['inputs'][0]['data_source_name']
        except Exception:
            click.secho("Data source must be provided in manifest or on commandline", fg='red')
            raise

    if not data_type:
        try:
            data_type = input_manifest['inputs'][0]['data_type_name']
        except Exception:
            click.secho("Data Type must be provided in manifest or on command line", fg='red')
            raise

    # image proc spec isn't always required.  if not given, check the manifest.  use it if it's provided.
    if not image_processing_spec:
        image_processing_spec = input_manifest['inputs'][0]['parameters'].get('image_processing_spec', None)
    data_parameters = {}
    if image_processing_spec:
        data_parameters['image_processing_spec'] = image_processing_spec

    click.echo("Using the following information:")
    click.echo(f"--> algorithm_version_id:     {algorithm_version_id}")
    click.echo(f"--> config_name:              {config_name}")
    click.echo(f"--> config_desc:              {config_desc}")
    click.echo(f"--> data_source:              {data_source}")
    click.echo(f"--> data_type:                {data_type}")
    click.echo(f"--> image_processing_spec:    {image_processing_spec}")

    if not dry_run:
        algorithm_config_id = asyncio.run(
            wf.create_algorithm_config(
                algorithm_version_id,
                config_name,
                config_desc,
                data_source,
                data_type,
                data_parameters,
            )
        )
        tsu.echo_highlight_suffix("You're new algorithm_config_id is: ", algorithm_config_id, 'green')


@config.command('deactivate')
@click.pass_context
@click.option('-ic', '--algorithm_config_ids', type=str, multiple=True, required=True,
              help="UUIDs of the algorithm_configs to deactivate")
def algorithm_config_deactivate(
        ctx,
        algorithm_config_ids,
):

    response = asyncio.run(wf.deactivate_algorithm_configs(algorithm_config_ids=algorithm_config_ids))

    click.echo("Deactivation response code: ", nl=False)
    if response:
        click.secho(f" {response} (FAILURE)", fg='red')
    else:
        click.secho(f" {response} (SUCCESS)", fg='green')



from datetime import datetime, timezone
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.struct_pb2 import Struct

from elements_api import ElementsAsyncClient
from elements_api.models.tasking_order_pb2 import (
    TaskingOrderCreateRequest, TaskingOrder, TimeRange, TaskingOrderApproveRequest, TaskingOrderCancelRequest,
    TaskingOrderGetRequest, TaskingOrderListRequest
)
from elements_api.models.order_pb2 import OrderState
from elements.sdk.tools.sdk_support import SDKSupport

class APITaskingOrder:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client
        self.sdk_support = SDKSupport()

    async def create(self, data_source_id: str, product_spec_name: str, target_geom_wkb: bytes, start_ts: datetime,
                     end_ts: datetime, metadata: Struct = None) -> TaskingOrder:
        request = TaskingOrderCreateRequest(
            data_source_id=data_source_id,
            product_spec_name=product_spec_name,
            target_geom=target_geom_wkb,
            acq_window=TimeRange(
                start_utc=Timestamp(seconds=int(start_ts.replace(tzinfo=timezone.utc).timestamp())),
                finish_utc=Timestamp(seconds=int(end_ts.replace(tzinfo=timezone.utc).timestamp()))
            ),
            metadata=metadata
        )
        response = await self.__client.api.tasking_order.create(request)
        return response.order

    async def approve(self, tasking_order_id: str, comment: str) -> None:
        request = TaskingOrderApproveRequest(
            id=tasking_order_id,
            comment=comment
        )
        await self.__client.api.tasking_order.approve(request)

    async def cancel(self, tasking_order_id: str) -> None:
        request = TaskingOrderCancelRequest(
            id=tasking_order_id
        )
        await self.__client.api.tasking_order.cancel(request)

    async def get(self, tasking_order_id: str) -> TaskingOrder:
        request = TaskingOrderGetRequest(
            id=tasking_order_id
        )
        response = await self.__client.api.tasking_order.get(request)
        return response.order

    async def list(self, data_source_ids: list[str], states: list[OrderState], analysis_computation_ids: list[str],
                   algorithm_computation_ids: list[str]) -> list[TaskingOrder]:
        request = TaskingOrderListRequest(
            data_source_ids=data_source_ids,
            states=states if states else [],
            analysis_computation_ids=analysis_computation_ids,
            algorithm_computation_ids=algorithm_computation_ids
        )
        responses = await self.sdk_support.get_all_paginated_objects(request,
                                                                     api_function=self.__client.api.tasking_order.list,
                                                                     timeout=self.__timeout)
        tasking_orders = []
        for response in responses:
            tasking_orders.extend(response.orders)
        return tasking_orders



#!/usr/bin/env python

import click
import asyncio
import uuid
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup


@click.command(cls=AliasedGroup, help="'visualization' command group")
@click.pass_context
def visualization(ctx):
    pass


@visualization.command('get')
@click.pass_context
@click.option('-nz', '--visualizer_config_name', type=str)
@click.option('-v', '--verbose', is_flag=True)
def visualization_get(ctx, visualizer_config_name, verbose):

    visualizer_config_id = str(uuid.uuid5(uuid.NAMESPACE_OID, visualizer_config_name))

    if verbose:
        pass
    click.secho("WARNING: This is a convenience function which computes the ID as ", fg='red')
    click.secho("         uuid.uuid5(visualizer_config_name).  If you misspelled the name, ", fg='red')
    click.secho("         then the ID will not actually exist in the system.", fg='red')
    tsu.echo_highlight_suffix(f"visualizer_config_id:  '{visualizer_config_name}' --> ",
                              visualizer_config_id, 'green')


@visualization.command('create')
@click.option('-nz', '--visualizer_config_names', multiple=True, type=str)
@click.option('-iv', '--algorithm_version_id', type=str)
@click.option('-v', '--verbose', is_flag=True)
@click.pass_context
def visualization_create(ctx, visualizer_config_names, algorithm_version_id, verbose=False):

    visualizer_config_ids = asyncio.run(
        wf.create_visualizer_config_algo(visualizer_config_names, algorithm_version_id)
    )
    tsu.echo_highlight_suffix("Your new visualizer_config_id is: ", visualizer_config_ids, 'green')



#!/usr/bin/env python

import click
import asyncio
import re
import yaml
import pandas as pd
import tabulate
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedSuperGroup, AliasedGroup


@click.command(cls=AliasedGroup, help='Data source command group')
@click.pass_context
def source(ctx):
    pass


@source.command('get')
@click.pass_context
@click.option('-n', '--source_name', type=str)
def data_source_get(ctx, source_name):

    wf.check_environment_complete(raise_on_failure=True)

    data_sources = asyncio.run(wf.list_data_sources())

    data = []
    for data_source in data_sources:
        data_source = tsu.protobuf_to_dict(data_source)
        if re.search(source_name, data_source['id']):
            data.append(data_source)
    df = pd.DataFrame(data)
    click.secho("Supported Data Sources:", fg='cyan')
    click.echo(tabulate.tabulate(df.T, maxcolwidths=[None, 40]))


@source.command('list')
@click.pass_context
def data_source_list(ctx):

    wf.check_environment_complete(raise_on_failure=True)

    data_sources = asyncio.run(wf.list_data_sources())

    data = []
    for data_source in data_sources:
        data.append(tsu.protobuf_to_dict(data_source))
    df = pd.DataFrame(data)
    pd.set_option('display.max_rows', None)
    click.secho("Supported Data Sources:", fg='cyan')
    click.echo(df[['id', 'name']])


@click.command(cls=AliasedGroup, help='Data type command group')
@click.pass_context
def type(ctx):
    pass


@type.command('get')
@click.pass_context
@click.option('-n', '--type_name', type=str)
def data_type_get(ctx, type_name):

    wf.check_environment_complete(raise_on_failure=True)

    data_types = asyncio.run(wf.list_data_types())

    data = []
    for data_type in data_types:
        data_type = tsu.protobuf_to_dict(data_type)
        if re.search(type_name, data_type['name']):
            data.append(data_type)
    df = pd.DataFrame(data)

    click.secho("Supported Data Types:", fg='cyan')
    click.echo(tabulate.tabulate(df.T, maxcolwidths=[None, 40]))


@type.command('list')
@click.pass_context
def data_type_list(ctx):

    wf.check_environment_complete(raise_on_failure=True)

    data_types = asyncio.run(wf.list_data_types())

    data = []
    for data_type in data_types:
        data.append(tsu.protobuf_to_dict(data_type))
    df = pd.DataFrame(data)
    pd.set_option('display.max_rows', None)
    click.secho("Supported Data Types:", fg='cyan')
    click.echo(df[['name', 'sensor_type', 'data_source_ids']])


@type.command('create')
@click.pass_context
@click.option('-f', '--definition_file', type=str, default=None)
@click.option('-n', '--name', type=str, default=None)
@click.option('-d', '--description', type=str, default=None)
@click.option('-s', '--schema', type=str, default=None)
@click.option('-ds', '--data_source_ids', default=None)
@click.option('-S', '--sensor_type', type=str, default=None)
def data_type_create(
        ctx,
        definition_file=None,
        name=None,
        description=None,
        schema=None,
        data_source_ids=None,
        sensor_type=''
):

    wf.check_environment_complete(raise_on_failure=True)

    if definition_file:
        with open(definition_file) as fp:
            data_type_def = yaml.safe_load(fp)
        if not name:
            name = data_type_def['name']
        if not description:
            description = data_type_def.get('description', '')
        if not schema:
            schema = data_type_def.get('schema', '')
        if not data_source_ids:
            data_source_ids = data_type_def.get('data_source_ids', '')
        if not sensor_type:
            sensor_type = data_type_def.get('sensor_type', '')

    data_type = asyncio.run(wf.create_data_type(name=name, description=description,
                                                schema=schema, data_source_ids=data_source_ids,
                                                sensor_type=sensor_type))
    click.echo(data_type)


@click.command(cls=AliasedSuperGroup, help="'data' super group")
@click.pass_context
def data(ctx):
    pass


data.add_command(source)
data.add_command(type)



from enum import Enum
from typing import List

from elements_api import ElementsAsyncClient
from elements_api.models.visualization_pb2 import (
    VisualizationGetRequest, Visualization, VisualizerConfigAlgoVersionCreateRequest, VisualizerConfigAlgoVersionCreateResponse,
    VisualizerConfigAlgoConfigCreateResponse, VisualizerConfigAlgoConfigCreateRequest
)


class VisualizationConfigType(Enum):
    UNKNOWN_TYPE = 0
    STANDARD = 1
    SPATIAL = 2
    TEMPORAL = 3


class APIVisualization:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def get(self, result_observation_ids: List) -> List[Visualization]:
        """

        :param result_observation_ids:
        :return: List[Visualization]
        """
        request = VisualizationGetRequest(
            result_observation_ids=result_observation_ids
        )
        response = await self.__client.api.visualization.get(request)
        return response.visualizations

    async def create_visualizer_config_algo_version(self, visualizer_config_id: str,
                                            algorithm_version_id: str) -> VisualizerConfigAlgoVersionCreateResponse:
        request = VisualizerConfigAlgoVersionCreateRequest(
            visualizer_config_id=visualizer_config_id,
            algorithm_version_id=algorithm_version_id
        )
        response = await self.__client.api.visualization.create_config_algo_version(request)
        return response

    async def create_visualizer_config_algo_config(self, visualizer_config_id: str,
                                                   algorithm_config_id: str) -> VisualizerConfigAlgoConfigCreateResponse:
        request = VisualizerConfigAlgoConfigCreateRequest(
            visualizer_config_id=visualizer_config_id,
            algorithm_config_id=algorithm_config_id
        )
        response = await self.__client.api.visualization.create_config_algo_config(request)
        return response



#!/usr/bin/env python

import click
import asyncio
import pandas as pd
import elements.cli.lib.workflow as wf
import elements.cli.lib.utils as tsu
from elements.cli.lib.aliased_group import AliasedGroup


@click.command(cls=AliasedGroup, help="'aoi' command group")
@click.pass_context
def aoi(ctx):
    pass


@aoi.command('get')
@click.pass_context
@click.option('-i', '--aoi_collection_id', type=str)
@click.option('-v', '--verbose', is_flag=True)
def aoi_get(ctx, aoi_collection_id, verbose):

    aoi_collection = asyncio.run(wf.get_aoi_collection(aoi_collection_id))
    data = []
    for aoi in aoi_collection:
        data.append(tsu.protobuf_to_dict(aoi))
    df = pd.DataFrame(data)

    if verbose:
        click.echo(aoi_collection)
    click.echo(df.T)


@aoi.command('list')
@click.pass_context
@click.option('-v', '--verbose', is_flag=True, default=False)
def aoi_list(ctx, verbose=False):

    aoi_collections = asyncio.run(wf.list_aoi_collections())
    aoi_collections = tsu.protobuf_to_dict(aoi_collections)['aoi_versions']
    data = []
    for aoi_collection in aoi_collections:
        data.append(aoi_collection)
    df = pd.DataFrame(data)

    if verbose:
        tsu.set_pandas_display()
    else:
        df = df[['aoi_id', 'aoi_name', 'timezone']]

    click.echo(df)


@aoi.command('create')
@click.pass_context
@click.option('-a', '--aoi_file', type=str)
def aoi_create(ctx, aoi_file):

    aoi_collection_id = asyncio.run(wf.create_aoi_collection(aoi_file))
    tsu.echo_highlight_suffix("Your new aoi_collection_id is: ", aoi_collection_id, 'green')


@aoi.command('update')
@click.pass_context
@click.option('-i', '--aoi_collection_id', type=str, required=True)
@click.option('-a', '--aoi_file', type=str)
def aoi_update(ctx, aoi_collection_id, aoi_file):

    aoi_collection_id = asyncio.run(wf.upload_aoi(aoi_collection_id, aoi_file))
    tsu.echo_highlight_suffix("Your updated aoi_collection_id is: ", aoi_collection_id, 'green')


@aoi.command('delete')
@click.pass_context
@click.option('-i', '--aoi_collection_ids', multiple=True)
def aoi_delete(ctx, aoi_collection_ids):

    click.secho("I'm sorry.  It is not possible to delete an aoi_collection", fg='red')
    # asyncio.run(wf.delete_aois(aoi_collection_ids))



import logging
import os
import re
import uuid
from datetime import datetime
import shapely.wkt as shp_wkt
import pandas as pd
from google.protobuf.timestamp_pb2 import Timestamp

import elements.cli.lib.utils as tsu
from elements.sdk.builder.algorithm import AlgorithmManifest, AlgorithmConfiguration
from elements.sdk.builder.toi import TOIBuilder, TOIRuleBuilder, Frequency
from elements.sdk.elements_sdk import ElementsSDK
from elements.sdk.builder.analysis import AnalysisManifest, AnalysisConfiguration
from elements_api.models.permission_pb2 import Permission
from elements_api.models.resource_pb2 import Resource
from elements_api import ElementsAsyncClient
from typing import List


VALUE_PRICE_DEFAULT = 0.01
EXECUTION_PRICE_DEFAULT = 0.01


############################################################################
# Init
############################################################################
def check_environment_complete(raise_on_failure=False, print_missing=True, verbose=False, requires_admin=False):
    """
    Check that the required environment parameters are present.

    This function checks whether the required environment parameters are present
    and meets the specified conditions.

    Args:
        raise_on_failure (bool, optional): If True, raise an exception if any required parameter is missing.
            Defaults to False.
        print_missing (bool, optional): If True, print the names of missing parameters to the console.
            Defaults to True.

    Returns:
        bool: True if all required parameters are present; False otherwise.

    Raises:
        EnvironmentError: Raised if raise_on_failure is True and any required parameter is missing.
    """

    missing = set()
    required_environment = ['ELEMENTS_HOST', 'ELEMENTS_TOKEN', 'ELEMENTS_AUTHOR']
    if requires_admin:
        required_environment.append('ELEMENTS_ADMIN_TOKEN')
    var_list = []
    for var in required_environment:
        value = os.getenv(var, None)
        var_list.append([var, value])
        if value is None:
            missing.add(var)
    df = pd.DataFrame(var_list, columns=['Variable', 'Value'])
    if verbose:
        print('')
        print(df)
        print('')

    if missing:
        msg = "The following required environment variables are missing:\n"
        for var in sorted(missing):
            msg += f"--> {var}"

        if print_missing:
            print(msg)
            print('')

        if raise_on_failure:
            raise EnvironmentError(f"Missing environment variables. [{','.join(missing)}]")

    return len(missing) == 0


############################################################################
# Manifest
############################################################################
def check_manifest_complete(manifest_dict, print_message=False):
    """
    Check if the provided manifest dictionary is complete.

    This function examines the given manifest dictionary to ensure that it contains
    all the required fields for proper functionality.

    Args:
        manifest_dict (dict): The dictionary representing the manifest.
            This dictionary should contain specific fields for proper validation.
        print_message (bool, optional): Whether to print a message about the completion status.
            Defaults to False.

    Returns:
        bool: True if the manifest is complete, False otherwise.
    """

    # This whole thing is ugly ... do this more nicely!

    required = {
        'container_parameters': ['image', 'command'],
        'metadata': ['description', ],
        'interface': ['interface_type'],
        'inputs': [['data_type_name']],
        'outputs': [['data_type_name']],
    }

    out_str = ''
    for param, subparams in required.items():
        if param not in manifest_dict:
            out_str += f'--> {param:40s} MISSING\n'
        else:
            out_str += f'--> {param:40s} OK\n'
            for subparam in subparams:
                # junky hack for inputs.data_type_name
                manifest_check_field = manifest_dict[param]
                if isinstance(subparam, list):
                    subparam = subparam[0]
                    manifest_check_field = manifest_check_field[0]

                if subparam not in manifest_check_field:
                    out_str += f'--> {param+"."+subparam:40s} MISSING\n'
                else:
                    out_str += f'--> {param+"."+subparam:40s} OK\n'
    if 'MISSING' in out_str:
        if print_message:
            print(out_str)
        return False
    return True


def update_manifest_docker_hash(manifest, docker_version_hash=None):
    """Update the Docker version hash in the given manifest dictionary.

    This function allows updating the Docker version hash in the provided manifest
    dictionary. If a new Docker version hash is provided, it will be assigned to
    the 'docker_hash' field in the manifest.

    Args:
        manifest (dict): The dictionary representing the manifest.
            This dictionary is expected to have a container_parameters.image field for updating.
        docker_version_hash (str, optional): The new Docker version hash to assign.
            If not provided, the 'docker_hash' field in the manifest will be cleared.

    Returns:
        image (str): The full image name of the docker image.

    Raises: RuntimeError: Raised if the input docker version hash is
        not present in the manifest, and is not provided as an input
        argument 'docker_version_hash'.
    """

    # replace the docker version hash in the manifest, if one is provided
    params = manifest['container_parameters']
    if docker_version_hash:
        image = params['image']
        image = image.rstrip(":.*") + f":{docker_version_hash}"
        params['image'] = image
    else:
        if ':' not in params['image']:
            raise RuntimeError('Missing docker hash. Not in manifest->container_parameters.image, and none provided')

    return params['image']


def create_algorithm_manifest(manifest_dict):
    """Create an algorithm Manifest object from a dictionary.

    Args:
        manifest_dict (dict): The dictionary representing the manifest.

    Returns:
        manifest (AlgorithmManifest): The manifest object.
    """

    if not check_manifest_complete(manifest_dict, print_message=True):
        raise RuntimeError("Manifest missing required parameters")

    container_parameters = manifest_dict['container_parameters']
    resource_request = container_parameters['resource_request']
    metadata = manifest_dict['metadata']
    interface = manifest_dict['interface']
    inputs = manifest_dict['inputs']
    outputs = manifest_dict['outputs']
    parallelizations = manifest_dict['parallelization']

    manifest = AlgorithmManifest()

    # metadata
    manifest.metadata_required(description=metadata['description'])
    manifest.metadata_tags(metadata['tags'])

    # container parameters
    manifest.container_parameters_required(image=container_parameters['image'],
                                           command=container_parameters['command'])
    manifest.container_parameters_resource_request(**resource_request)

    # interface
    manifest.interface_required(interface_type=interface['interface_type'])

    # inputs/outputs
    for inp in inputs:
        data_type_name = inp['data_type_name']
        data_type_params = {k: v for k, v in inp.items() if k != 'data_type_name'}
        manifest.inputs_add_data_type(data_type_name=data_type_name, **data_type_params)

    for output in outputs:
        manifest.outputs_add_data_type(**output)

    for parallelization in parallelizations:
        manifest.parallelization_add_configuration(parallelization)

    # manifest.grouping_frequency()

    # parameters
    algorithm_parameters = manifest_dict.get('parameters', [])
    for parameter in algorithm_parameters:
        manifest.parameter_add(**parameter)

    return manifest


def create_analysis_manifest(
        name: str,
        algorithm_version_id: str,
        description: str,
        tags: List,
):
    """
    Create an analysis manifest for a specific algorithm version.

    This function creates an analysis manifest based on the provided information
    about the analysis, algorithm version, and associated details.

    Args:
        name (str): The name of the analysis.
        algorithm_version_id (str): The ID of the algorithm version used for the analysis.
        description (str): A description of the analysis.
        tags (List[str]): A list of tags associated with the analysis.

    Returns:
        manifest (AnalysisManifest): The analysis manifest containing provided information.
    """

    manifest = AnalysisManifest()
    manifest.metadata(description=description, tags=tags)
    manifest.add_node(name, algorithm_version_id)

    return manifest


############################################################################
#  Algorithm
############################################################################
async def get_algorithms(algorithm_ids: List = None):
    """Retrieve information about an algorithm by name.

    This asynchronous function retrieves information about an algorithm.

    Args:
        algorithm_ids (List): The IDs of the algorithms to retrieve.

    Returns:
        List[Algorithm]: A list of Algorithm objects.
    """

    algorithms = []
    sdk = ElementsSDK()
    if algorithm_ids is not None:
        algorithms += await sdk.algorithm.get(algorithm_ids=algorithm_ids)
    return algorithms


async def list_algorithms(algorithm_name: str = None):
    """Retrieve information about an algorithm by name.

    This asynchronous function retrieves information about an algorithm, optionally
    filtered by the specified algorithm name.

    Args:
        algorithm_name (str, optional): The name of the algorithm to filter results.
            If provided, only information about the specified algorithm will be retrieved.

    Returns:
        List[Algorithm]: A list of Algorithm objects.
    """

    kwargs = {}
    sdk = ElementsSDK()
    if algorithm_name is not None:
        kwargs['search_text'] = algorithm_name
    algorithms = await sdk.algorithm.list(**kwargs)
    return algorithms


async def get_algorithm_versions(algorithm_id: str = None, algorithm_version_id: str = None):
    """
    Retrieve algorithm versions for a specified algorithm_id.

    This asynchronous function retrieves algorithm_versions of a specific algorithm,
    based on its algorithm ID.

    Args:
        algorithm_id (str): The ID of the algorithm for which to retrieve version information.

    Returns:
        List[AlgorithmVersion]: A list of algorithm versions.
    """

    algorithm_versions = []
    sdk = ElementsSDK()
    if algorithm_id:
        algorithm_versions += await sdk.algorithm_version.list(algorithm_id=algorithm_id)
    if algorithm_version_id:
        algorithm_versions += await sdk.algorithm_version.get(algorithm_version_ids=[algorithm_version_id])

    return algorithm_versions


async def get_current_algorithm_version(algorithm_id: str):
    """Get the next algorithm version
    """

    algo_versions = await get_algorithm_versions(algorithm_id=algorithm_id)
    if algo_versions:
        data = []
        for algo_version in algo_versions:
            algo_version_dict = tsu.protobuf_to_dict(algo_version)
            data.append(algo_version_dict)

        df = pd.DataFrame(data=data)
        current_version = df['version'].max()
    else:
        current_version = None
    return current_version


async def create_algorithm(name: str, author: str, display_name: str):
    """
    Create a new algorithm.

    This asynchronous function creates a new algorithm with the provided information,
    including the algorithm's name, author, and display name.

    Args:
        name (str): The internal name of the algorithm.
        author (str): The author's name or identifier.
        display_name (str): The display name of the algorithm.

    Returns:
        id(str):  The algorithm ID for the newly created algorithm,
    """
    sdk = ElementsSDK()
    algorithm = await sdk.algorithm.create(name=name, author=author,
                                           display_name=display_name)
    logging.info("\"algorithm_id\": \"{}\"".format(algorithm.id))
    return algorithm.id


async def create_algorithm_version(
        algorithm_id: str,
        manifest,
):
    """
    Create a new algorithm version for a specified algorithm_id.

    This asynchronous function creates a new version for an existing algorithm
    with the provided algorithm ID and manifest.

    Args:
        algorithm_id (str): The ID of the algorithm for which to create a new version.
        manifest (AlgorithmManifest): The manifest containing details.

    Returns:
        algorithm_version_id (str): The algorithm_version_id of the newly created algorithm version.
    """

    sdk = ElementsSDK()
    algorithm_version = await sdk.algorithm_version.create(algorithm_id=algorithm_id,
                                                           algorithm_manifest=manifest)

    logging.info("\"algorithm_version_id\": \"{}\"".format(algorithm_version.id))
    logging.info("algorithm_sem-version: {}".format(algorithm_version.version))
    return algorithm_version


def get_elements_envvars():
    elements_host = os.environ["ELEMENTS_HOST"]
    elements_port_string = os.getenv("ELEMENTS_PORT", default="443")
    elements_port = int(elements_port_string)
    return elements_host, elements_port


async def deactivate_algorithm_versions(
        algorithm_version_ids: List[str],
):
    """
    """
    elements_host, elements_port = get_elements_envvars()
    secure_string = os.getenv("ELEMENTS_SECURE", default="True")
    secure = True if secure_string in ["True", "true", "Yes", "yes", "1"] else False
    # admin_api_token = os.environ["ELEMENTS_ADMIN_TOKEN"]

    client = ElementsAsyncClient(elements_host, elements_port, secure=secure)
    # admin_client = ElementsAsyncClient(oi_papi_host, oi_papi_port, secure=secure, api_token=admin_api_token)

    deactivate_algorithm_versions_request = client.models.algorithm_version_pb2.AlgorithmVersionDeactivateRequest(
        ids=algorithm_version_ids
    )
    deactivate_algo_versions_response = await client.api.algorithm_version.deactivate(
        deactivate_algorithm_versions_request
    )

    return deactivate_algo_versions_response.status_code


async def update_algorithm(
        algorithm_id: str,
        manifest,
        value_price=VALUE_PRICE_DEFAULT,
        execution_price=EXECUTION_PRICE_DEFAULT,
        visualizer_config_names=None,
):
    """
    Update an existing algorithm's information.

    This asynchronous function updates the information of an existing algorithm
    with the provided algorithm ID and manifest. It also allows updating the value
    and execution prices, with default values if not provided.

    Args:
        algorithm_id (str): The ID of the algorithm to update.
        manifest (dict): The updated manifest dictionary with new information.
        value_price (float, optional): The new value price for the algorithm.
            Defaults to VALUE_PRICE_DEFAULT if not provided.
        execution_price (float, optional): The new execution price for the algorithm.
            Defaults to EXECUTION_PRICE_DEFAULT if not provided.
        visualizer_config_names (List, optional): Names of visualizations for this algorithm.

    Returns:
        algorithm_version_id (str): The algorithm_version_id for the algorithm.
    """
    algorithm_version = await create_algorithm_version(algorithm_id, manifest)

    # disable credit setting for now, as it requires an admin token
    # await set_credit(algorithm_version_id=algorithm_version.id, value_price=value_price, execution_price=execution_price)

    visualizer_config_ids = []
    if visualizer_config_names:
        visualizer_config_ids = await create_visualizer_config_algo_version(visualizer_config_names,
                                                                            algorithm_version.id)

    logging.info("\"algorithm_version_id\": \"{}\"".format(algorithm_version.id))
    logging.info("algorithm_sem-version: {}".format(algorithm_version.version))
    logging.info("visualizer_config_names: {}".format(visualizer_config_names))
    logging.info("visualizer_config_ids: {}".format(visualizer_config_ids))
    return algorithm_version.id


async def new_algorithm(
        name,
        author,
        display_name,
        manifest,
        value_price=VALUE_PRICE_DEFAULT,
        execution_price=EXECUTION_PRICE_DEFAULT,
        visualizer_config_names=None,
):
    """
    Create a new algorithm with initial information.

    This asynchronous function creates a new algorithm with the provided information,
    including the algorithm's name, author, display name, manifest, and optional
    value and execution prices.

    Args:
        name (str): The internal name of the algorithm.
        author (str): The author's name or identifier.
        display_name (str): The display name of the algorithm.
        manifest (AlgorithmManifest): The initial manifest object containing version and details.
        value_price (float, optional): The initial value price for the algorithm.
            Defaults to VALUE_PRICE_DEFAULT if not provided.
        execution_price (float, optional): The initial execution price for the algorithm.
            Defaults to EXECUTION_PRICE_DEFAULT if not provided.
        visualizer_config_names (List, optional): Names of visualizations for this algorithm.

    Returns:
        algorithm_version_id (str): The algorithm_version_id.
    """
    algorithm_id = await create_algorithm(name, author, display_name)
    algorithm_version_id = await update_algorithm(algorithm_id,
                                                  manifest,
                                                  value_price=value_price,
                                                  execution_price=execution_price,
                                                  visualizer_config_names=visualizer_config_names)

    return algorithm_id, algorithm_version_id


async def get_algorithm_configs(algorithm_version_id: str = None, algorithm_config_id: str = None):
    """
    Retrieve algorithm configs for a specific algorithm version.

    This asynchronous function retrieves configuration details for a specific algorithm version
    identified by the provided algorithm version ID.

    Args:
        algorithm_version_id (str): The ID of the algorithm version for which to retrieve associated configs.
        algorithm_config_id (str): The ID of the algorithm config to retrieve.

    Returns:
        list(AlgorithmConfig): A list of AlgorithmConfig objects for the specified algorithm version.
    """

    sdk = ElementsSDK()
    algorithm_configs = []
    if algorithm_version_id:
        algorithm_configs += await sdk.algorithm_config.list(algorithm_version_id=algorithm_version_id)
    if algorithm_config_id:
        algorithm_configs += await sdk.algorithm_config.get(ids=[algorithm_config_id])

    return algorithm_configs


async def create_algorithm_config(
        algorithm_version_id: str,
        config_name: str,
        config_desc: str,
        data_source,
        data_type: str,
        data_parameters: dict = None,
        visualizer_config_names: list = None,
        **kwargs,
):
    """
    Create a new algorithm config for a specific algorithm version.

    This asynchronous function creates a new algorithm config for a specific algorithm version
    identified by the provided algorithm version ID. The configuration details include the
    configuration name, description, data source, and optional data type.

    Args:
        algorithm_version_id (str): The ID of the algorithm version to which the config belongs.
        config_name (str): The name of the configuration.
        config_desc (str): A description of the configuration.
        data_source: The data source for the configuration.
        data_type (DataType, optional): The data type associated with the configuration.
        visualizer_config_names (list, optional): Names associated with visualizer configurations.

    Returns:
        algorithm_config_id (str): The ID of the newly created algorithm config.
    """
    sdk = ElementsSDK()

    if data_source is not None and data_source.value is not None:
        kwargs['data_source'] = data_source

    # Config
    config = AlgorithmConfiguration()
    config.add_data_source(data_type=data_type,
                           data_parameters=data_parameters, **kwargs)

    algorithm_config = await sdk.algorithm_config.create(
        algorithm_version_id=algorithm_version_id,
        name=config_name + f"_{str(uuid.uuid4())[:8]}",
        description=config_desc,
        algorithm_config=config
    )

    visualizer_config_ids = []
    if visualizer_config_names:
        visualizer_config_ids = await create_visualizer_config_algo_config(visualizer_config_names,
                                                                           algorithm_config.id)
    logging.info("\"algorithm_config_id\": \"{}\"".format(algorithm_config.id))
    logging.info("\"visualizer_config_names\": \"{}\"".format(visualizer_config_names))
    logging.info("\"visualizer_config_ids\": \"{}\"".format(visualizer_config_ids))

    return algorithm_config.id


async def deactivate_algorithm_configs(
        algorithm_config_ids: List[str]
):
    elements_host, elements_port = get_elements_envvars()
    secure_string = os.getenv("ELEMENTS_SECURE", default="True")
    secure = True if secure_string in ["True", "true", "Yes", "yes", "1"] else False
    # admin_api_token = os.environ["ELEMENTS_ADMIN_TOKEN"]

    client = ElementsAsyncClient(elements_host, elements_port, secure=secure)
    # admin_client = ElementsAsyncClient(oi_papi_host, oi_papi_port, secure=secure, api_token=admin_api_token)

    deactivate_algorithm_configs_request = client.models.algorithm_config_pb2.AlgorithmConfigDeactivateRequest(
        ids=algorithm_config_ids
    )
    deactivate_algorithm_configs_response = await client.api.algorithm_config.deactivate(
        deactivate_algorithm_configs_request
    )
    return deactivate_algorithm_configs_response.status_code


############################################################################
#  Analysis
############################################################################
async def get_analyses(analysis_ids: List):
    """
    Retrieve analyses for specified analysis IDs

    This asynchronous function retrieves analysis objects by ID.

    Args:
        analysis_ids (List): The analysis_ids to retrieve.

    Returns:
        List[Analysis]: A list of Analysis objects.
    """
    sdk = ElementsSDK()
    analyses = await sdk.analysis.get(ids=analysis_ids)
    return analyses


async def list_analyses(analysis_name: str = None):
    """
    Retrieve analyses, optionally filtered by name.

    This asynchronous function retrieves analysis objects, optionally
    filtered by the specified analysis name.

    Args:
        analysis_name (str, optional): The name of the analysis to filter results.
            If provided, only information about the specified analysis will be retrieved.

    Returns:
        List[Analysis]: A list of Analysis objects.
    """
    kwargs = {}
    sdk = ElementsSDK()
    analyses = await sdk.analysis.list(**kwargs)

    # Do not use 'search_text' in kwargs, as that doesn't allow regular expressions
    # just get everything and use regex here
    if analysis_name:
        analyses_matched = []
        for analysis in analyses:
            match_name = re.search(analysis_name, analysis.name)
            match_auth = re.search(analysis_name, analysis.author)
            match_id = re.search(analysis_name, analysis.id)
            if match_name or match_auth or match_id:
                analyses_matched.append(analysis)
        analyses = analyses_matched

    return analyses


async def get_analysis_versions(analysis_version_id: str = None, analysis_id: str = None):
    """
    Retrieve analysis_versions for a specific analysis_id.

    This asynchronous function retrieves information about versions of a specific analysis
    identified by the provided analysis ID.

    Args:
        analysis_id (str): The ID of the analysis for which to retrieve version information.

    Returns:
        List[AnalysisVersion]: A list of analysis versions.
    """
    sdk = ElementsSDK()
    analysis_versions = []
    if analysis_id:
        analysis_versions += await sdk.analysis_version.list(analysis_id=analysis_id,
                                                             include_algorithm_details=True, include_manifest=True)
    if analysis_version_id:
        analysis_versions += await sdk.analysis_version.get(ids=[analysis_version_id],
                                                            include_algorithm_details=True, include_manifest=True)

    return analysis_versions


async def get_current_analysis_version(analysis_id: str):
    """Get the current analysis version
    """

    analysis_versions = await get_analysis_versions(analysis_id=analysis_id)
    if analysis_versions:
        data = []
        for analysis_version in analysis_versions:
            analysis_version_dict = tsu.protobuf_to_dict(analysis_version)
            data.append(analysis_version_dict)

        df = pd.DataFrame(data=data)
        current_version = df['analysis_manifest.metadata.version'].max()
    else:
        current_version = None
    return current_version


async def create_analysis(name: str, author: str):
    """
    Create a new analysis.

    This asynchronous function creates a new analysis with the provided name and author.

    Args:
        name (str): The name of the new analysis.
        author (str): The author's name or identifier.

    Returns:
        analysis_id (str):  The ID of the new analysis.
    """
    sdk = ElementsSDK()
    analysis = await sdk.analysis.create(name=name, author=author)

    logging.info("\"analysis_id\": \"{}\"".format(analysis.id))
    return analysis.id


async def create_analysis_version(analysis_id: str, manifest):
    """
    Create a new analysis version for an existing analysis.

    This asynchronous function creates a new analysis version for an existing analysis
    identified by the provided analysis ID. The new version is defined by the
    provided manifest.

    Args:
        analysis_id (str): The ID of the analysis for which to create a new version.
        manifest (dict): The manifest dictionary containing version and details.

    Returns:
        analysis_version_id (str): The newly created analysis version ID.
    """
    sdk = ElementsSDK()
    analysis_version = await sdk.analysis_version.create(analysis_id=analysis_id,
                                                         analysis_manifest=manifest)

    return analysis_version.id


async def update_analysis(
        analysis_id: str,
        manifest,
        value_price=VALUE_PRICE_DEFAULT,
        execution_price=EXECUTION_PRICE_DEFAULT
):
    """
    Update an existing analysis's information.

    This asynchronous function updates the information of an existing analysis
    identified by the provided analysis ID. It also allows updating the value
    and execution prices, with default values if not provided.

    Args:
        analysis_id (str): The ID of the analysis to update.
        manifest (dict): The updated manifest dictionary with new information.
        value_price (float, optional): The new value price for the analysis.
            Defaults to VALUE_PRICE_DEFAULT if not provided.
        execution_price (float, optional): The new execution price for the analysis.
            Defaults to EXECUTION_PRICE_DEFAULT if not provided.

    Returns:
        analysis_version_id (str):  The analysis_version_id of the updated analysis.
    """

    analysis_version = await create_analysis_version(analysis_id, manifest)
    # disable set credit for now, as it requires an admin token
    # await set_credit(analysis_version_id=analysis_version.id, value_price=value_price,
    #                  execution_price=execution_price)
    logging.info("\"analysis_version_id\": \"{}\"".format(analysis_version.id))
    logging.info("analysis_sem-version: {}".format(analysis_version.version))
    return analysis_version.id


async def new_analysis(
        name,
        author,
        display_name,
        manifest,
        value_price=VALUE_PRICE_DEFAULT,
        execution_price=EXECUTION_PRICE_DEFAULT
):
    """
    Create a new analysis with initial information.

    This asynchronous function creates a new analysis with the provided information,
    including the analysis's name, author, display name, manifest, and optional
    value and execution prices.

    Args:
        name (str): The internal name of the analysis.
        author (str): The author's name or identifier.
        display_name (str): The display name of the analysis.
        manifest (Manifest): The initial manifest containing version and details.
        value_price (float, optional): The initial value price for the analysis.
            Defaults to VALUE_PRICE_DEFAULT if not provided.
        execution_price (float, optional): The initial execution price for the analysis.
            Defaults to EXECUTION_PRICE_DEFAULT if not provided.

    Returns:
        analysis_version_id (str): The Analysis Version ID of the newly created analysis,
    """

    analysis_id = await create_analysis(name, author, display_name)
    analysis_version_id = await update_analysis(analysis_id,
                                                manifest,
                                                value_price=value_price,
                                                execution_price=execution_price)

    return analysis_version_id


async def get_analysis_configs(analysis_config_id: str = None, analysis_version_id: str = None):
    """
    Retrieve analysis configs for a specific analysis version.

    This asynchronous function retrieves analysis config details for a specific analysis version
    identified by the provided analysis version ID.

    Args:
        analysis_version_id (str): The ID of the analysis version for which to retrieve configurations.

    Returns:
        list(AnalysisConfig): A list of analysis configs for the specified analysis version ID.
    """

    sdk = ElementsSDK()
    analysis_configs = []
    if analysis_config_id:
        analysis_configs += await sdk.analysis_config.get(ids=[analysis_config_id])
    if analysis_version_id:
        analysis_configs += await sdk.analysis_config.list(analysis_version_id=analysis_version_id,
                                                           include_deactivated=True)

    return analysis_configs


async def list_analysis_configs(analysis_config_name: str = None):
    """
    Retrieve analysis configs for a analysis by name / regex

    Args:
        analysis_version_name (str): The name or regex of the analysis config for which to retrieve configurations.

    Returns:
        list(AnalysisConfig): A list of analysis configs for the specified analysis version ID.
    """

    sdk = ElementsSDK()
    analysis_configs = []
    analysis_configs += await sdk.analysis_config.list(search_text=analysis_config_name)

    return analysis_configs


async def create_analysis_config(
        analysis_version_id: str,
        algorithm_config_id: str,
        analysis_config_name: str,
        analysis_config_desc: str,
):
    """
    Create a new analysis config for a specific analysis version.

    This asynchronous function creates a new analysis config for a specific analysis version,
    identified by the provided analysis version ID. The configuration details include the
    analysis configuration name, description, associated algorithm configuration ID, and name.

    Args:
        analysis_version_id (str): The ID of the analysis version to which the configuration belongs.
        algorithm_config_id (str): The ID of the associated algorithm configuration.
        analyais_config_name (str): The name of the associated analysis configuration.
        analysis_config_desc (str): A description of the analysis configuration.

    Returns:
        analysis_config_id (str): analysis config Id for the newly created analysis config.
    """

    sdk = ElementsSDK()

    analysis = await sdk.analysis_version.get(ids=[analysis_version_id],
                                              include_algorithm_details=True, include_manifest=True)
    node_name = analysis[0].analysis_manifest.algorithm_nodes[0].name

    # Config
    config = AnalysisConfiguration(analysis_version_id=analysis_version_id)
    config.add_config_node(
        name=node_name,
        algorithm_config_id=algorithm_config_id,
    )

    analysis_config = await sdk.analysis_config.create(
        analysis_version_id=analysis_version_id,
        algorithm_config_nodes=config.get(),
        name=analysis_config_name,
        description=analysis_config_desc,
    )

    logging.info("\"analysis_config_id\": \"{}\"".format(analysis_config.id))

    return analysis_config.id


async def deactivate_analysis_configs(
        analysis_config_ids: List[str]
):
    elements_host, elements_port = get_elements_envvars()
    secure_string = os.getenv("ELEMENTS_SECURE", default="True")
    secure = True if secure_string in ["True", "true", "Yes", "yes", "1"] else False
    # admin_api_token = os.environ["ELEMENTS_ADMIN_TOKEN"]

    client = ElementsAsyncClient(elements_host, elements_port, secure=secure)
    # admin_client = ElementsAsyncClient(oi_papi_host, oi_papi_port, secure=secure, api_token=admin_api_token)

    deactivate_analysis_configs_request = client.models.analysis_config_pb2.AnalysisConfigDeactivateRequest(
        ids=analysis_config_ids
    )
    deactivate_analysis_configs_response = await client.api.analysis_config.deactivate(
        deactivate_analysis_configs_request
    )
    return deactivate_analysis_configs_response.status_code


############################################################################
# Price
############################################################################
async def get_credit(credit_source_id: str):
    """
    Retrieve the credit information for a specific credit source.

    This asynchronous function retrieves the price information associated with a specific
    credit source identified by the provided credit source ID.

    Args:
        credit_source_id (str): The ID of the credit source for which to retrieve price information.

    Returns:
        summary: A credit summary for the specified credit source.
    """
    sdk = ElementsSDK()
    summary = await sdk.credit.summary(credit_source_id)
    return summary


async def set_credit(
        algorithm_version_id: str,
        value_price=VALUE_PRICE_DEFAULT,
        execution_price=EXECUTION_PRICE_DEFAULT
):
    """
    Set the credit information for a specific algorithm version.

    This asynchronous function sets the price information for a specific algorithm version
    identified by the provided algorithm version ID. It allows setting both the value price
    and the execution price, with default values if not provided.

    In order this to work, the user must also have the following in their environment.

    export ELEMENTS_ADMIN_TOKEN=<admin_token>

    Args:
        algorithm_version_id (str): The ID of the algorithm version for which to set the price information.
        value_price (float, optional): The new value price for the algorithm version.
            Defaults to VALUE_PRICE_DEFAULT if not provided.
        execution_price (float, optional): The new execution price for the algorithm version.
            Defaults to EXECUTION_PRICE_DEFAULT if not provided.

    Returns:
        None

    """
    elements_host, _ = get_elements_envvars()
    terrascope_api_admin_token = os.getenv('ELEMENTS_ADMIN_TOKEN')
    assert terrascope_api_admin_token is not None
    client = ElementsSDK().create_client(terrascope_host=elements_host,
                                         terrascope_api_token=terrascope_api_admin_token)
    sdk = ElementsSDK(client=client)
    await sdk.credit.set_algorithm(algorithm_version_id=algorithm_version_id,
                                   algorithm_value_price=value_price,
                                   algorithm_execution_price=execution_price)


############################################################################
# TOI
############################################################################
async def create_toi(
        start: str,
        end: str,
        date_format: str = "%Y-%m-%dT%H:%M:%SZ",
        frequency=Frequency.DAILY,
):
    """
    Create a new time of interest (TOI) object.

    This asynchronous function creates a new time of interest (TOI) object based on the provided
    start and end times, along with optional parameters for date format and frequency.

    Args:
        start (str): The start time of the time of interest (TOI) in the specified date format.
        end (str): The end time of the time of interest (TOI) in the specified date format.
        date_format (str, optional): The date format of the start and end times.
            Defaults to "%Y-%m-%dT%H:%M:%SZ" if not provided.
        frequency (Frequency, optional): The frequency of the time of interest (TOI).
            Defaults to Frequency.DAILY if not provided.

    Returns:
        toi_id (str): The toi_id of the newly created time of interest (TOI) object,
    """
    sdk = ElementsSDK()

    toi_configuration = TOIBuilder()
    toi_configuration.build_toi(start=datetime.strptime(start, date_format),
                                finish=datetime.strptime(end, date_format))
    toi_configuration.build_recurrence(TOIRuleBuilder.build_rule(
        frequency=Frequency.DAILY,
        interval=1))
    toi = await sdk.toi.create(toi_configuration.get())
    logging.info("\"toi_id\": \"{}\"".format(toi.id))
    return toi.id


async def get_tois(toi_ids: List):
    """
    Retrieve information about multiple time of interest (TOI) objects.

    This asynchronous function retrieves information about multiple time of interest (TOI) objects
    identified by the provided list of TOI IDs.

    Args:
        toi_ids (List): A list of TOI IDs for which to retrieve information.

    Returns:
        List[Toi]: A list of Toi objects for the requested TOIs.
    """
    sdk = ElementsSDK()
    if toi_ids:
        tois = await sdk.toi.get(ids=toi_ids)
    else:
        tois = await sdk.toi.list()
    return tois


async def delete_tois(toi_ids: List):
    """
    Delete multiple time of interest (TOI) objects.

    This asynchronous function deletes multiple time of interest (TOI) objects identified by
    the provided list of TOI IDs.

    Args:
        toi_ids (List): A list of TOI IDs for the TOIs to be deleted.

    Returns:
        None
    """
    sdk = ElementsSDK()
    await sdk.toi.delete(ids=toi_ids)


############################################################################
# AOI
############################################################################
async def upload_aoi(aoi_collection_id, file_path):
    """
    Upload an area of interest (AOI) file to a collection.

    This asynchronous function uploads an area of interest (AOI) file specified by the provided
    file path to the given AOI collection identified by its ID.  Only geojson is currently supported.

    Args:
        aoi_collection_id (str): The ID of the AOI collection to which the file will be uploaded.
        file_path (str): The file path to the area of interest (AOI) file to be uploaded.
                         Only geojson is currently supported.

    Returns:
        response: What is this?
    """

    sdk = ElementsSDK()
    response = await sdk.aoi.upload(aoi_collection_id=aoi_collection_id,
                                    file_path=file_path)
    return response


async def create_aoi_collection(file_path: str):
    """
    Create a new area of interest (AOI) collection using a file.

    This asynchronous function creates a new area of interest (AOI) collection based on the
    data in the file specified by the provided file path.  Only geojson is currently supported.

    Args:
        file_path (str): The file path to the data file used to create the AOI collection.
                         Only geojson is currently supported.

    Returns:
        aoi_collection_id (str):  The ID of the newly created aoi collection.
    """

    sdk = ElementsSDK()
    aoi_collection = await sdk.aoi_collection.create(aoi_collection_name=str(uuid.uuid4())[:8])
    await sdk.aoi.upload(aoi_collection_id=aoi_collection.id,
                         file_path=file_path)
    logging.info("\"aoi_collection_id\": \"{}\"".format(aoi_collection.id))
    return aoi_collection.id


async def get_aoi_collection(aoi_collection_id: str):
    """
    Retrieve information about an area of interest (AOI) collection.

    This asynchronous function retrieves information about an area of interest (AOI) collection
    identified by the provided AOI collection ID.

    Args:
        aoi_collection_id (str): An AOI collection ID for which to retrieve information.

    Returns:
        aoi_collection: The requested AOI collection.

    """
    sdk = ElementsSDK()
    if aoi_collection_id is not None:
        aoi_collections = [await sdk.aoi_collection.get(aoi_collection_id)]
    else:
        aoi_collections = await sdk.aoi_collection.list()
    return aoi_collections


async def list_aoi_collections():
    """
    Retrieve information about all area of interest (AOI) collections.

    This asynchronous function retrieves information about area of interest (AOI) collections

    Returns:
        aoi_collections: The requested AOI collection.

    """
    sdk = ElementsSDK()
    aoi_collections = await sdk.aoi_collection.list()
    return aoi_collections


async def delete_aois(aoi_ids: List):
    """
    Delete multiple area of interest (TOI) objects.

    This asynchronous function deletes multiple area of interest (TOI) objects identified by
    the provided list of AOI IDs.

    Args:
        aoi_ids (List): A list of AOI IDs for the AOIs to be deleted.

    Returns:
        None
    """
    sdk = ElementsSDK()
    await sdk.aoi.delete(aoi_collection_ids=aoi_ids)


############################################################################
# Algorithm Computation
############################################################################
async def create_algorithm_computation(
        algorithm_config_id: str,
        toi_id: str,
        aoi_collection_id: str
):
    """
    Create a new algorithm computation using provided algorithm_config_id, TOI, and AOI collection.

    This asynchronous function creates a new algorithm computation using the provided algorithm configuration,
    time of interest (TOI) ID, and area of interest (AOI) collection ID.

    Args:
        algorithm_config_id (str): The ID of the algorithm configuration to be used in the computation.
        toi_id (str): The ID of the time of interest (TOI) to be used in the computation.
        aoi_collection_id (str): The ID of the area of interest (AOI) collection to be used in the computation.

    Returns:
        algorithm_computation_id (str):  The ID of the newly created algorithm computation.
    """
    sdk = ElementsSDK()

    algorithm_computation = await sdk.algorithm_computation.create(
        aoi_collection_id=aoi_collection_id,
        toi_id=toi_id,
        algorithm_config_id=algorithm_config_id
    )

    logging.info("\"algorithm_computation_id\": \"{}\"".format(algorithm_computation))
    return algorithm_computation.id


async def get_algorithm_computation_info(algorithm_computation_ids: List, verbose=False):
    """
    Retrieve information about multiple algorithm computations.

    This asynchronous function retrieves information about multiple algorithm computations identified by
    the provided list of algorithm computation IDs. The verbosity parameter can be used to control
    the level of detail in the returned information.

    Args:
        algorithm_computation_ids (List): A list of algorithm computation IDs for which to retrieve information.
        verbose (bool, optional): If True, detailed information will be printed.
            Defaults to False.

    Returns:
        List[computation]: A list of the requested algorithm computations.
    """
    sdk = ElementsSDK()
    algorithm_computations = await sdk.algorithm_computation.get(computation_ids=algorithm_computation_ids)

    for algorithm_computation in algorithm_computations:
        if verbose:
            print("Algorithm Computation: ")
            logging.info(algorithm_computation)
        algorithm_config = await sdk.algorithm_config.get(ids=[algorithm_computation.algo_config_id])
        if verbose:
            print("Algorithm Config: ")
            logging.info(algorithm_config)

    return algorithm_computations


async def run_algorithm_computations(algorithm_computation_ids: List):
    """
    Run algorithm computations for a list of algorithm computation IDs.

    This asynchronous function initiates the execution of algorithm computations for the
    provided list of algorithm computation IDs. The computations will be run asynchronously.

    Args:
        algorithm_computation_ids (List): A list of algorithm computation IDs to run.

    Returns:
        None
    """
    sdk = ElementsSDK()
    await sdk.algorithm_computation.run(algorithm_computation_ids=algorithm_computation_ids)

    return None


async def download_algorithm_computation_results(
    algorithm_computation_ids: List[str] = None, analysis_computation_ids: List[str] = None,
    source_aoi_version: int = None, observation_start_ts: str = None, max_observation_start_ts: str = None,
    download_dir: str = None
):
    """
    Download results from a completed algorithm computation.

    This asynchronous function downloads the results from a completed algorithm computation
    identified by the provided algorithm computation ID. The results will be saved to the specified
    file path, with a default name of "test.zip" if not provided.

    Args:
        algorithm_computation_ids (List[str]): The IDs of the completed algorithm computations.
        analysis_computation_ids (List[str]): The IDs of the completed analysis computations

    Returns:
        Dict[str, Dict[str, str]: mapping of algorithm_computation_id_to_data_type_to_merged_file
    """
    sdk = ElementsSDK()
    min_observation_start_timestamp = None
    if observation_start_ts:
        min_observation_start_dt = datetime.strptime(observation_start_ts, '%Y-%m-%d')
        min_observation_start_timestamp = Timestamp()
        min_observation_start_timestamp.FromDatetime(min_observation_start_dt)
    max_observation_start_timestamp = None
    if max_observation_start_ts:
        max_observation_start_dt = datetime.strptime(max_observation_start_ts, '%Y-%m-%d')
        max_observation_start_timestamp = Timestamp()
        max_observation_start_timestamp.FromDatetime(max_observation_start_dt)

    algorithm_computation_id_to_data_type_to_downloaded_paths = await sdk.result.download(
        algorithm_computation_ids=algorithm_computation_ids, analysis_computation_ids=analysis_computation_ids,
        source_aoi_version=source_aoi_version, observation_start_ts=min_observation_start_timestamp,
        max_observation_start_ts=max_observation_start_timestamp, download_dir=download_dir
    )
    algorithm_computation_id_to_data_type_to_merged_file = await sdk.result.merge_download_files(
        algorithm_computation_id_to_data_type_to_downloaded_paths, download_dir
    )
    return algorithm_computation_id_to_data_type_to_merged_file


############################################################################
# Analysis Computation
############################################################################
async def create_analysis_computation(
        analysis_config_id: str,
        toi_id: str,
        aoi_collection_id: str
):
    """
    Create a new analysis computation using provided analysis config, TOI, and AOI collection.

    This asynchronous function creates a new analysis computation using the provided analysis configuration,
    time of interest (TOI) ID, and area of interest (AOI) collection ID.

    Args:
        analysis_config_id (str): The ID of the analysis configuration to be used in the computation.
        toi_id (str): The ID of the time of interest (TOI) to be used in the computation.
        aoi_collection_id (str): The ID of the area of interest (AOI) collection to be used in the computation.

    Returns:
        analysis_computation_id (str):  The analysis_computation_id of the newly created analysis computation.
    """
    sdk = ElementsSDK()

    analysis_computation = await sdk.analysis_computation.create(
        aoi_collection_id=aoi_collection_id,
        toi_id=toi_id,
        analysis_config_id=analysis_config_id
    )

    logging.info("\"analysis_computation_id\": \"{}\"".format(analysis_computation))
    return analysis_computation.id


async def get_analysis_computation_info(
        min_created_on=None,
        max_created_on=None,
        verbose=False,
        search_text=None,
):
    """
    Retrieve information about multiple analysis computations.

    This asynchronous function retrieves information about multiple analysis computations.
    The verbosity parameter can be used to control the level of detail in the returned information.

    Args:
        verbose (bool, optional): If True, detailed information will be included in the results.
            Defaults to False.

    Returns:
        min_created_on (date): Minimum created_on date to query
        max_created_on (date): Maximum created_on date to query
        List[Computations]: A list of computations.
    """

    kwargs = {}
    enabled = False  # date specs are currently not working in sdk.
    if min_created_on and enabled:
        kwargs['min_created_on'] = min_created_on
    if max_created_on and enabled:
        kwargs['max_created_on'] = max_created_on
    if search_text:
        kwargs['search_text'] = search_text

    sdk = ElementsSDK()
    computations = await sdk.analysis_computation.list(**kwargs)

    for computation in computations:
        if verbose:
            print("Analysis Computation: ")
            logging.info(computation)

    return computations


async def run_analysis_computations(analysis_computation_ids: List):
    """
    Run analysis computations for a list of analysis computation IDs.

    This asynchronous function initiates the execution of analysis computations for the
    provided list of analysis computation IDs. The computations will be run asynchronously.

    Args:
        analysis_computation_ids (List): A list of analysis computation IDs to run.

    Returns:
        None
    """
    sdk = ElementsSDK()
    await sdk.analysis_computation.run(analysis_computation_ids=analysis_computation_ids)

    return None


############################################################################
# Permissions
############################################################################
async def get_permissions(
        analysis_config_id,
):
    """
    Retrieve permissions associated with an analysis configuration.

    This asynchronous function retrieves the permissions associated with a specific analysis configuration
    identified by the provided analysis configuration ID.

    Args:
        analysis_config_id: The ID of the analysis configuration for which to retrieve permissions.

    Returns:
        List[permission]: A list of the permissions associated with
            the specified analysis configuration.
    """
    sdk = ElementsSDK()
    permissions = await sdk.permission.get(analysis_config_ids=[analysis_config_id])
    return permissions


async def resource_permission_create(
        resource_ids: List,
        user_emails: List,
        permission_type=Permission.Type.READ,
        resource_type=Resource.Type.ALGORITHM,
        public=False,
        public_confirm=False,
):
    """
    Grant resource permissions for specified users.

    This asynchronous function creates resource permissions for the specified users, granting them
    the specified permission type on the specified resource IDs. The permissions can be associated
    with different resource types, and there's an option to set the permissions as public.

    In order this to work, the user must also have the following in their environment.

    export ELEMENTS_ADMIN_TOKEN=<admin_token>

    Args:
        resource_ids (List): A list of resource IDs for which to grant permissions.
        user_emails (List): A list of user emails to whom the permissions will be granted.
        permission_type (Permission.Type, optional): The type of permission to grant.
            Defaults to Permission.Type.READ.
        resource_type (Resource.Type, optional): The type of resource for which to grant permissions.
            Defaults to Resource.Type.ALGORITHM.
        public (bool, optional): If True, set the permissions as public. Defaults to False.
        public_confirm (bool, optional): If True, confirm setting permissions as public. Defaults to False.

    Returns:
        None
    """

    if 0:  # what is going on here??
        elements_host, _ = get_elements_envvars()
        terrascope_api_admin_token = os.getenv('ELEMENTS_ADMIN_TOKEN')
        assert terrascope_api_admin_token is not None
        client = ElementsSDK().create_client(terrascope_host=terrascope_host,
                                             terrascope_api_token=terrascope_api_admin_token)
        sdk = ElementsSDK(client=client)
    else:
        sdk = ElementsSDK()

    resp = await sdk.permission.create(resource_ids=resource_ids,
                                       permission_type=permission_type,
                                       resource_type=resource_type,
                                       user_ids=user_emails,
                                       public=public,
                                       public_confirm=public_confirm)
    logging.info(f"Status code: {resp.status_code}  public={public}")


async def analysis_permission_create(
        analysis_config_id: str,
        user_emails: List,
        permission_type=Permission.Type.READ,
        public=False,
        public_confirm=False,
        chunk_size=10
):
    """
    Grant analysis configuration permissions for specified users.

    This asynchronous function creates analysis configuration permissions for the specified users,
    granting them the specified permission type on the specified analysis configuration. The permissions
    can be set as public, and the chunk size parameter determines how many permissions are created in
    each batch.

    Args:
        analysis_config_id (str): The ID of the analysis configuration for which to grant permissions.
        user_emails (List): A list of user emails to whom the permissions will be granted.
        permission_type (Permission.Type, optional): The type of permission to grant.
            Defaults to Permission.Type.READ.
        public (bool, optional): If True, set the permissions as public. Defaults to False.
        public_confirm (bool, optional): If True, confirm setting permissions as public. Defaults to False.
        chunk_size (int, optional): The number of permissions to create in each batch.
            Defaults to 10.

    Returns:
        None
    """
    # get the algorithm, algorithm_config, analysis, and analysis_config_id
    sdk = ElementsSDK()
    ac = await sdk.analysis_config.get([analysis_config_id], include_algorithm_details=True)
    ac = ac[0]
    algorithms = [f.algorithm.id for f in ac.algorithm_configs]
    algorithm_configs = [f.id for f in ac.algorithm_configs]
    analyses = [ac.analysis_version.analysis.id]

    await resource_permission_create(algorithms, user_emails,
                                     permission_type, Resource.Type.ALGORITHM,
                                     public=public, public_confirm=public_confirm)
    await resource_permission_create(analyses, user_emails,
                                     permission_type, Resource.Type.ANALYSIS,
                                     public=public, public_confirm=public_confirm)
    for algo_configs in tsu.chunk_array(algorithm_configs, chunk_size):
        await resource_permission_create(algo_configs, user_emails,
                                         permission_type, Resource.Type.ALGORITHM_CONFIG,
                                         public=public, public_confirm=public_confirm)
    await resource_permission_create([analysis_config_id], user_emails,
                                     permission_type, Resource.Type.ANALYSIS_CONFIG,
                                     public=public, public_confirm=public_confirm)


############################################################################
# Data
############################################################################
async def list_data_sources():
    """
    Get list of supported data sources.

    """
    sdk = ElementsSDK()
    data_sources = await sdk.data_source.list()
    return data_sources


async def create_data_type(
        name: str,
        description: str,
        schema: str,
        data_source_ids: List,
        sensor_type: str
):
    """
    Create a new data type.
    """

    sdk = ElementsSDK()
    await sdk.data_type.create(
        name=name,
        description=description,
        schema=schema,
        data_source_ids=data_source_ids,
        sensor_type=sensor_type,
    )


async def list_data_types():
    """
    Get list of supported data types.

    """
    sdk = ElementsSDK()
    data_types = await sdk.data_type.list()
    return data_types


############################################################################
# Imagery
############################################################################
async def search_imagery(
        geometry_wkt,
        datetime_start,
        datetime_end,
        data_source_id,
        product_spec_name,
        search_service='SCENE'):
    """
    """
    wkb = shp_wkt.loads(geometry_wkt).wkb
    sdk = ElementsSDK()
    scenes = await sdk.imagery.search(
        wkb,
        datetime_start,
        datetime_end,
        data_source_id,
        product_spec_name,
        search_service=search_service
    )
    return scenes


############################################################################
# Visualization
############################################################################

def get_visualizer_config_uuid(visualizer_config_name):

    # This will return the same uuid for the same string.
    visualizer_config_id = str(uuid.uuid5(uuid.NAMESPACE_OID, visualizer_config_name))
    return visualizer_config_id


async def create_visualizer_config_algo_version(
        visualizer_config_names: List,
        algorithm_version_id: str
):

    sdk = ElementsSDK()

    visualizer_config_ids = []
    for visualizer_config_name in visualizer_config_names:

        visualizer_config_id = get_visualizer_config_uuid(visualizer_config_name)
        await sdk.visualization.create_visualizer_config_algo_version(visualizer_config_id=visualizer_config_id,
                                                                      algorithm_version_id=algorithm_version_id)
        logging.info(f"Set Visualizer Config: {visualizer_config_name} = {visualizer_config_id}")
        visualizer_config_ids.append(visualizer_config_id)

    return visualizer_config_ids


async def create_visualizer_config_algo_config(
        visualizer_config_names: List,
        algorithm_config_id: str
):

    sdk = ElementsSDK()

    visualizer_config_ids = []
    for visualizer_config_name in visualizer_config_names:

        visualizer_config_id = get_visualizer_config_uuid(visualizer_config_name)
        await sdk.visualization.create_visualizer_config_algo_config(visualizer_config_id=visualizer_config_id,
                                                                     algorithm_config_id=algorithm_config_id)
        logging.info(f"Set Visualizer Config: {visualizer_config_name} = {visualizer_config_id}")
        visualizer_config_ids.append(visualizer_config_id)

    return visualizer_config_ids


############################################################################
# Multi-function
############################################################################
async def create_algorithm_computation_aoi_toi(
        algorithm_config_id,
        start_date,
        end_date,
        aoi_file_path,
        date_format="%Y-%m-%d",
        frequency=Frequency.DAILY,
):
    """
    Create an algorithm computation with AOI and TOI from specified parameters.

    This asynchronous function creates an algorithm computation using the provided algorithm configuration,
    start and end dates for the time of interest (TOI), an AOI file path, and optional parameters for date format
    and frequency.

    Args:
        algorithm_config_id: The ID of the algorithm configuration to be used in the computation.
        start_date: The start date of the time of interest (TOI) in the specified date format.
        end_date: The end date of the time of interest (TOI) in the specified date format.
        aoi_file_path: The file path to the area of interest (AOI) file to be used in the computation.
        date_format (str, optional): The date format of the start and end dates. Defaults to "%Y-%m-%d".
        frequency (Frequency, optional): The frequency of the time of interest (TOI). Defaults to Frequency.DAILY.

    Returns:
        algorithm_computation_id (str): The Algorithm computation ID of the newly created algorithm computation.
    """
    sdk = ElementsSDK()

    toi_id = await create_toi(
        start=start_date,
        end=end_date,
        date_format=date_format,
        frequency=frequency,
    )
    aoi_collection_id = await create_aoi_collection(file_path=aoi_file_path)

    algorithm_computation = await sdk.algorithm_computation.create(
        aoi_collection_id=aoi_collection_id,
        toi_id=toi_id,
        algorithm_config_id=algorithm_config_id
    )
    logging.info("----------------------\n")
    return algorithm_computation.id


async def create_analysis_computation_aoi_toi(
        analysis_config_id,
        start_date,
        end_date,
        aoi_file_path,
        date_format="%Y-%m-%d",
        frequency=Frequency.DAILY,
):
    """
    Create an analysis computation with AOI and TOI from specified parameters.

    This asynchronous function creates an analysis computation using the provided analysis configuration,
    start and end dates for the time of interest (TOI), an AOI file path, and optional parameters for date format
    and frequency.

    Args:
        analysis_config_id: The ID of the analysis configuration to be used in the computation.
        start_date: The start date of the time of interest (TOI) in the specified date format.
        end_date: The end date of the time of interest (TOI) in the specified date format.
        aoi_file_path: The file path to the area of interest (AOI) file to be used in the computation.
        date_format (str, optional): The date format of the start and end dates. Defaults to "%Y-%m-%d".
        frequency (Frequency, optional): The frequency of the time of interest (TOI). Defaults to Frequency.DAILY.

    Returns:
        analysis_computation_id (str):  The analysis computation ID of the newly created analysis computation.
    """
    sdk = ElementsSDK()

    toi_id = await create_toi(
        start=start_date,
        end=end_date,
        date_format=date_format,
        frequency=frequency,
    )
    aoi_collection_id = await create_aoi_collection(file_path=aoi_file_path)

    analysis_computation = await sdk.analysis_computation.create(
        aoi_collection_id=aoi_collection_id,
        toi_id=toi_id,
        analysis_config_id=analysis_config_id
    )
    logging.info("----------------------\n")
    return analysis_computation.id


async def create_and_run_algorithm_computation(
        algorithm_config_id,
        start_date,
        end_date,
        aoi_file_path,
        date_format="%Y-%m-%d",
        frequency=Frequency.DAILY,
):
    """
    Create and run an algorithm computation with AOI and TOI from specified parameters.

    This asynchronous function combines the creation and execution of an algorithm computation using
    the provided algorithm configuration, start and end dates for the time of interest (TOI), an AOI file path,
    and optional parameters for date format and frequency.

    Args:
        algorithm_config_id: The ID of the algorithm configuration to be used in the computation.
        start_date: The start date of the time of interest (TOI) in the specified date format.
        end_date: The end date of the time of interest (TOI) in the specified date format.
        aoi_file_path: The file path to the area of interest (AOI) file to be used in the computation.
        date_format (str, optional): The date format of the start and end dates. Defaults to "%Y-%m-%d".
        frequency (Frequency, optional): The frequency of the time of interest (TOI). Defaults to Frequency.DAILY.

    Returns:
        algorithm_computation_id (str):  The algorithm computation ID of the newly created and (hopefully)
                                         running algorithm computation.
    """
    sdk = ElementsSDK()

    algorithm_computation_id = await create_algorithm_computation_aoi_toi(
        algorithm_config_id,
        start_date,
        end_date,
        aoi_file_path,
        date_format=date_format,
        frequency=frequency,
    )

    await sdk.algorithm_computation.run(
        algorithm_computation_ids=[algorithm_computation_id])

    return algorithm_computation_id


async def create_and_run_analysis_computation(
        analysis_config_id,
        start_date,
        end_date,
        aoi_file_path,
        date_format="%Y-%m-%d",
        frequency=Frequency.DAILY,
):
    """
    Create and run an analysis computation with AOI and TOI from specified parameters.

    This asynchronous function combines the creation and execution of an analysis computation using
    the provided analysis configuration, start and end dates for the time of interest (TOI), an AOI file path,
    and optional parameters for date format and frequency.

    Args:
        analysis_config_id: The ID of the analysis configuration to be used in the computation.
        start_date: The start date of the time of interest (TOI) in the specified date format.
        end_date: The end date of the time of interest (TOI) in the specified date format.
        aoi_file_path: The file path to the area of interest (AOI) file to be used in the computation.
        date_format (str, optional): The date format of the start and end dates. Defaults to "%Y-%m-%d".
        frequency (Frequency, optional): The frequency of the time of interest (TOI). Defaults to Frequency.DAILY.

    Returns:
        analysis_computation_id (str): The analysis computation ID of the newly created and (hopefully)
                                       running analysis computation.
    """
    sdk = ElementsSDK()

    analysis_computation_id = await create_analysis_computation_aoi_toi(
        analysis_config_id,
        start_date,
        end_date,
        aoi_file_path,
        date_format=date_format,
        frequency=frequency,
    )

    await sdk.analysis_computation.run(
        analysis_computation_ids=[analysis_computation_id])

    return analysis_computation_id



from google.protobuf.timestamp_pb2 import Timestamp
from elements_api import ElementsAsyncClient
from elements_api.models.common_models_pb2 import Pagination
from elements_api.models.credit_pb2 import Credit as PapiCredit, CreditEstimateRequest, CreditTransactionsResponse
from elements_api.models.credit_pb2 import CreditAddRequest, CreditRemoveRequest, CreditRefundRequest, \
    CreditAlgorithmMultiplierSetRequest, CreditDataSourceMultiplierSetRequest, CreditSummaryRequest, \
    CreditTransactionsRequest, Transaction


class APICredit:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def estimate(self, algorithm_computation_id: str) -> float:
        """
        Description:
        Returns an estimate of the credits required to compute results for the specified algorithm.
        Estimates the full cost, including open orders based on the ending date of the specified TOI.

        :return: APICredit Estimate : float
        """
        request = CreditEstimateRequest(
            algorithm_computation_id=algorithm_computation_id
        )
        response = await self.__client.api.credit.estimate(request, timeout=self.__timeout)
        return response.credit_estimate

    async def add(self, credit_source_id: str, amount: float, reason: str):
        """
        Description
        Add credits to the specified credit_source.

        :param credit_source_id: str
        :param amount: float
        :param reason: str
        :return: No Return
        """
        request = CreditAddRequest(
            credit_source_id=credit_source_id,
            amount=amount,
            reason=reason
        )
        await self.__client.api.credit.add(request, timeout=self.__timeout)

    async def remove(self, credit_source_id: str, amount: float, reason: str):
        """
        Description
        Removes credits to the specified credit_source.

        :param credit_source_id: str
        :param amount: float
        :param reason: str
        :return: No Return
        """
        request = CreditRemoveRequest(
            credit_source_id=credit_source_id,
            amount=amount,
            reason=reason
        )
        await self.__client.api.credit.remove(request, timeout=self.__timeout)

    async def refund(self, credit_source_id: str, amount: float, reason: str):
        """
        Description

        Refund credits to a user collection

        :param credit_source_id: str
        :param amount: float
        :param reason: str
        :return:
        """
        request = CreditRefundRequest(
            credit_source_id=credit_source_id,
            amount=amount,
            reason=reason
        )
        await self.__client.api.credit.refund(request, timeout=self.__timeout)

    async def summary(self, credit_source_id: str) -> PapiCredit:
        """
        Retrieves the credit summary for the specified user's department. As credits are associated with a department,
        all users of a department see the same credit summary. Currently, this can only be done at
        the department level.

        Available - Credits that can be used for future Analyses and Computations
        Reserved - Credits on hold for running Analyses and Computations
        Used - Credits used by past Analyses and Computations

        :param credit_source_id: str
        :return: APICredit
        """
        request = CreditSummaryRequest(
            credit_source_id=credit_source_id
        )
        response = await self.__client.api.credit.summary(request, timeout=self.__timeout)
        return response.credit

    async def transactions(self,
                           credit_source_id,
                           transaction_type: Transaction.TransactionType,
                           start_date: Timestamp,
                           end_date: Timestamp,
                           algorithm_computation_id: str,
                           pagination: Pagination) -> CreditTransactionsResponse:
        """
        Description

        Retrieves the detailed view of all transactions associated with a department. Can filter for a time range or
        transaction type. Currently, this can only be done at the department level.

        :param credit_source_id: str
        :param transaction_type: TransactionType
        :param start_date: Timestamp
        :param end_date: Timestamp
        :param algorithm_computation_id: str
        :param pagination: Pagination
        :return:
        """
        request = CreditTransactionsRequest(
            credit_source_id=credit_source_id,
            transaction_type=transaction_type,
            start_date=start_date,
            end_date=end_date,
            algorithm_computation_id=algorithm_computation_id,
            pagination=pagination
        )
        response = await self.__client.api.credit.transactions(request, timeout=self.__timeout)
        return response

    async def set_algorithm(self, algorithm_version_id: str,
                            algorithm_execution_price: float,
                            algorithm_value_price: float):
        """
        Description

        Assign a price to the specified algorithm. Price must be a multiple of $0.001.
        This can only be done by an admin user.

        :param algorithm_version_id: str
        :param algorithm_execution_price: float
        :param algorithm_value_price: float
        :return:
        """
        request = CreditAlgorithmMultiplierSetRequest(
            algorithm_version_id=algorithm_version_id,
            algorithm_execution_price=algorithm_execution_price,
            algorithm_value_price=algorithm_value_price
        )
        await self.__client.api.credit.set_algorithm(request, timeout=self.__timeout)

    async def set_data_source(self, data_source_id: str, data_source_price: float):
        """
        Description

        Assign a price to the specified data_source. Price must be a multiple of $0.001.
        This can only be done by an admin user.

        :param data_source_id: str
        :param data_source_price: float
        :return:
        """
        request = CreditDataSourceMultiplierSetRequest(
            data_source_id=data_source_id,
            data_source_price=data_source_price
        )
        await self.__client.api.credit.set_data_source(request, timeout=self.__timeout)



from typing import List

from elements_api import ElementsAsyncClient
from elements_api.models.user_pb2 import (
    LoginRequest, UserCreateRequest, User, UserGetRequest, UserListRequest, UserPublicKeyRequest, UserPublicKeyResponse
)


class APIUser:
    def __init__(self, client: ElementsAsyncClient, timeout):
        self.__timeout = timeout
        self.__client = client

    async def create(self, name: str, email: str, user_collection_id: str) -> User:
        """
        Description

        Create a new user in the system, assigned to the specified department. This functionality is only available
        for an admin user. This automatically creates an aoi_collection associated to this user.

        :param name:
        :param user_collection_id:
        :param email:
        :return:
        """
        request = UserCreateRequest(
            name=name,
            email=email,
            user_collection_id=user_collection_id
        )
        response = await self.__client.api.user.create(request)
        return response.user

    async def update(self, user_id: str, name: str = None, email: str = None):
        """
        Description

        Update Mutable fields for the given user.

        :param user_id:
        :param name:
        :param email:
        :return:
        """
        pass

    async def get(self, ids: List) -> List[User]:
        """
        Description
        Returns the details of the specified user. If no user is given, returns details for the requesting user.
        The specified user must belong to the same department as the requesting user, otherwise an error is thrown.

        :param ids:
        :return: List[User]
        """
        request = UserGetRequest(
            ids=ids
        )
        response = await self.__client.api.user.get(request)
        return response.users

    async def list(self, email: str) -> List[User]:
        """
        Description
        Returns the details of the specified user. If no user is given, returns details for the requesting user.
        The specified user must belong to the same department as the requesting user, otherwise an error is thrown.

        :param email:
        :return: List[User]
        """
        request = UserListRequest(
            email=email
        )
        response = await self.__client.api.user.list(request)
        return response.users

    async def delete(self, user_ids: List):
        """
        Description

        Delete a user from the system, ensuring that they are no longer able to log into the system and they are
        not able to consume any credits. This functionality is only available for an admin user.
        This is a soft delete, so references to the user do not break.

        :param user_ids:
        :return:
        """
        pass

    async def login(self, email: str, password: str):
        """
        Description

        Login using user's email and password. The endpoint returns a JWT access token that is needed to use the API.
        Note: This endpoint does NOT require token authentication.

        :param email:
        :param password:
        :return:
        """
        request = LoginRequest(
            email=email,
            password=password
        )
        response = await self.__client.api.user.login(request, timeout=self.__timeout)
        return response

    async def public_key_request(self):
        request = UserPublicKeyRequest()
        response = await self.__client.api.user.get_public_key(request, timeout=self.__timeout)
        return response.key_pem

    async def send_temp_password(self, email: str):
        """
        Description

        Generates a temporary password that is sent to the user's email address, which can be used to regain access
        to the system. It is strongly encouraged to change your password after getting a temporary password.
        This endpoint does NOT require token authentication.

        :param email:
        :return:
        """

        pass

    async def change_password(self, user_id: str, old_password: str, new_password: str):
        """
        Description

        Changes the user's password. Although an authenticated endpoint, still requires the user to
        enter their old password in order to change it.

        :param user_id:
        :param old_password:
        :param new_password:
        :return:
        """

        pass
